from __future__ import annotations

from typing import TYPE_CHECKING, Any

from livecheck.special.package import extract_project, get_latest_package, is_package
import pytest

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

test_cases = {
    'test_extract_project_npmjs_org': {
        'url': 'https://registry.npmjs.org/package-name',
        'expected': ('registry.npmjs.org', 'package-name'),
        'is_package': True
    },
    'test_extract_project_yarnpkg_com': {
        'url': 'https://registry.yarnpkg.com/package-name',
        'expected': ('registry.yarnpkg.com', 'package-name'),
        'is_package': True
    },
    'test_extract_project_scoped_package': {
        'url': 'https://registry.npmjs.org/@scope/package-name',
        'expected': ('registry.npmjs.org', '@scope/package-name'),
        'is_package': True
    },
    'test_extract_project_invalid_url': {
        'url': 'https://registry.npmjs.org/',
        'expected': ('', ''),
        'is_package': False
    },
    'test_extract_project_invalid_url2': {
        'url': 'https://aregistry.npmjs.org/package-name',
        'expected': ('', ''),
        'is_package': False
    },
    'test_extract_project_invalid_url3': {
        'url': 'https://registry.npmjs.com/package-name',
        'expected': ('', ''),
        'is_package': False
    },
    'test_extract_project_empty_string': {
        'url': '',
        'expected': ('', ''),
        'is_package': False
    },
    'devcontainers': {
        'url': 'https://registry.npmjs.org/@devcontainers/cli/-/cli-devcontainer-0.72.0.tgz',
        'expected': ('registry.npmjs.org', '@devcontainers/cli'),
        'is_package': True
    },
}


@pytest.mark.parametrize('test_case', test_cases.values(), ids=test_cases.keys())
def test_extract_project(test_case: dict[str, Any]) -> None:
    assert extract_project(test_case['url']) == test_case['expected']


@pytest.mark.parametrize('test_case', test_cases.values(), ids=test_cases.keys())
def test_is_package(test_case: dict[str, Any]) -> None:
    assert is_package(test_case['url']) == test_case['is_package']


def make_mock_response(versions: Any) -> Any:
    class MockResponse:
        def json(self) -> Any:
            return {'versions': versions}

    return MockResponse()


def test_get_latest_package_success(mocker: MockerFixture) -> None:
    src_uri = 'https://registry.npmjs.org/package-name'
    ebuild = 'package-name-1.0.0.ebuild'
    settings = mocker.Mock()
    mocker.patch('livecheck.special.package.extract_project',
                 return_value=('registry.npmjs.org', 'package-name'))
    mock_get_content = mocker.patch('livecheck.special.package.get_content',
                                    return_value=make_mock_response(['1.0.0', '2.0.0']))
    mock_get_last_version = mocker.patch('livecheck.special.package.get_last_version',
                                         return_value={'version': '2.0.0'})
    result = get_latest_package(src_uri, ebuild, settings)
    assert result == '2.0.0'
    mock_get_content.assert_called_once()
    mock_get_last_version.assert_called_once()


def test_get_latest_package_no_versions(mocker: MockerFixture) -> None:
    src_uri = 'https://registry.npmjs.org/package-name'
    ebuild = 'package-name-1.0.0.ebuild'
    settings = mocker.Mock()
    mocker.patch('livecheck.special.package.extract_project',
                 return_value=('registry.npmjs.org', 'package-name'))
    mocker.patch('livecheck.special.package.get_content', return_value=make_mock_response([]))
    mocker.patch('livecheck.special.package.get_last_version', return_value=None)
    result = get_latest_package(src_uri, ebuild, settings)
    assert not result


def test_get_latest_package_invalid_url(mocker: MockerFixture) -> None:
    src_uri = 'https://invalid.url'
    ebuild = 'package-name-1.0.0.ebuild'
    settings = mocker.Mock()
    mocker.patch('livecheck.special.package.extract_project', return_value=('', ''))
    result = get_latest_package(src_uri, ebuild, settings)
    assert not result


def test_get_latest_package_get_content_returns_none(mocker: MockerFixture) -> None:
    src_uri = 'https://registry.npmjs.org/package-name'
    ebuild = 'package-name-1.0.0.ebuild'
    settings = mocker.Mock()
    mocker.patch('livecheck.special.package.extract_project',
                 return_value=('registry.npmjs.org', 'package-name'))
    mocker.patch('livecheck.special.package.get_content', return_value=None)
    result = get_latest_package(src_uri, ebuild, settings)
    assert not result



from __future__ import annotations

from typing import TYPE_CHECKING

from livecheck.special.checksum import get_latest_checksum_package, update_checksum_metadata

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_get_latest_checksum_package_match_and_checksum_mismatch(mocker: MockerFixture) -> None:
    url = 'https://example.com/foo-1.0.tar.gz'
    ebuild = 'cat/foo/foo-1.0.ebuild'
    repo_root = '/repo'
    manifest_content = ('DIST foo-1.0.tar.gz 1234 BLAKE2B deadbeef SHA512 cafebabe\n'
                        'DIST bar-2.0.tar.gz 5678 BLAKE2B 123456 SHA512 789abc\n')
    mocker.patch('livecheck.special.checksum.catpkg_catpkgsplit',
                 return_value=('cat/foo', 'foo', 'r0', '1.0'))
    mock_open = mocker.mock_open(read_data=manifest_content)
    mocker.patch('pathlib.Path.open', mock_open)
    mocker.patch('livecheck.special.checksum.hash_url', return_value=('beefdead', 'beefcafe', 1234))
    mocker.patch('livecheck.special.checksum.get_last_modified',
                 return_value='2024-06-01T12:00:00Z')
    version, last_modified, returned_url = get_latest_checksum_package(url, ebuild, repo_root)
    assert version == '1.0'
    assert last_modified == '2024-06-01T12:00:00Z'
    assert returned_url == url


def test_get_latest_checksum_package_match_and_checksum_match(mocker: MockerFixture) -> None:
    url = 'https://example.com/foo-1.0.tar.gz'
    ebuild = 'cat/foo/foo-1.0.ebuild'
    repo_root = '/repo'
    manifest_content = ('DIST foo-1.0.tar.gz 1234 BLAKE2B deadbeef SHA512 cafebabe\n')

    mocker.patch('livecheck.special.checksum.catpkg_catpkgsplit',
                 return_value=('cat/foo', 'foo', 'r0', '1.0'))
    mock_open = mocker.mock_open(read_data=manifest_content)
    mocker.patch('pathlib.Path.open', mock_open)
    mocker.patch('livecheck.special.checksum.hash_url', return_value=('deadbeef', 'cafebabe', 1234))
    result = get_latest_checksum_package(url, ebuild, repo_root)
    assert result == ('', '', '')


def test_get_latest_checksum_package_no_match(mocker: MockerFixture) -> None:
    url = 'https://example.com/foo-1.0.tar.gz'
    ebuild = 'cat/foo/foo-1.0.ebuild'
    repo_root = '/repo'
    manifest_content = ('DIST bar-2.0.tar.gz 5678 BLAKE2B 123456 SHA512 789abc\n')
    mocker.patch('livecheck.special.checksum.catpkg_catpkgsplit',
                 return_value=('cat/foo', 'foo', 'r0', '1.0'))
    mock_open = mocker.mock_open(read_data=manifest_content)
    mocker.patch('pathlib.Path.open', mock_open)
    result = get_latest_checksum_package(url, ebuild, repo_root)
    assert result == ('', '', '')


def test_update_checksum_metadata_updates_matching_line(mocker: MockerFixture) -> None:
    ebuild = 'cat/foo/foo-1.0.ebuild'
    url = 'https://example.com/foo-1.0.tar.gz'
    repo_root = '/repo'
    manifest_content = ('DIST foo-1.0.tar.gz 1234 BLAKE2B deadbeef SHA512 cafebabe\n'
                        'DIST bar-2.0.tar.gz 5678 BLAKE2B 123456 SHA512 789abc\n')
    expected_content = ('DIST foo-1.0.tar.gz 4321 BLAKE2B newbeef SHA512 newcafe\n'
                        'DIST bar-2.0.tar.gz 5678 BLAKE2B 123456 SHA512 789abc\n')
    mocker.patch('livecheck.special.checksum.catpkg_catpkgsplit',
                 return_value=('cat/foo', 'foo', 'r0', '1.0'))
    mocker.patch('livecheck.special.checksum.hash_url', return_value=('newbeef', 'newcafe', 4321))
    temp_file_mock = mocker.MagicMock()
    tf_mock = mocker.mock_open()
    temp_file_mock.open = tf_mock
    mocker.patch('livecheck.special.checksum.EbuildTempFile', return_value=temp_file_mock)
    temp_file_mock.__enter__.return_value = temp_file_mock
    temp_file_mock.__exit__.return_value = None
    manifest_open_mock = mocker.mock_open(read_data=manifest_content)
    mocker.patch('pathlib.Path.open', manifest_open_mock)
    update_checksum_metadata(ebuild, url, repo_root)
    handle = tf_mock()
    written_lines = ''.join(call.args[0] for call in handle.write.call_args_list)
    assert written_lines == expected_content


def test_update_checksum_metadata_no_matching_line(mocker: MockerFixture) -> None:
    ebuild = 'cat/foo/foo-1.0.ebuild'
    url = 'https://example.com/foo-1.0.tar.gz'
    repo_root = '/repo'
    manifest_content = ('DIST bar-2.0.tar.gz 5678 BLAKE2B 123456 SHA512 789abc\n')
    expected_content = manifest_content
    mocker.patch('livecheck.special.checksum.catpkg_catpkgsplit',
                 return_value=('cat/foo', 'foo', 'r0', '1.0'))
    mocker.patch('livecheck.special.checksum.hash_url', return_value=('newbeef', 'newcafe', 4321))
    temp_file_mock = mocker.MagicMock()
    tf_mock = mocker.mock_open()
    temp_file_mock.open = tf_mock
    mocker.patch('livecheck.special.checksum.EbuildTempFile', return_value=temp_file_mock)
    temp_file_mock.__enter__.return_value = temp_file_mock
    temp_file_mock.__exit__.return_value = None
    manifest_open_mock = mocker.mock_open(read_data=manifest_content)
    mocker.patch('pathlib.Path.open', manifest_open_mock)
    update_checksum_metadata(ebuild, url, repo_root)
    handle = tf_mock()
    written_lines = ''.join(call.args[0] for call in handle.write.call_args_list)
    assert written_lines == expected_content


def test_update_checksum_metadata_multiple_matching_lines(mocker: MockerFixture) -> None:
    ebuild = 'cat/foo/foo-1.0.ebuild'
    url = 'https://example.com/foo-1.0.tar.gz'
    repo_root = '/repo'
    manifest_content = ('DIST foo-1.0.tar.gz 1234 BLAKE2B deadbeef SHA512 cafebabe\n'
                        'DIST foo-1.0.tar.gz 1234 BLAKE2B deadbeef SHA512 cafebabe\n')
    expected_content = ('DIST foo-1.0.tar.gz 4321 BLAKE2B newbeef SHA512 newcafe\n'
                        'DIST foo-1.0.tar.gz 4321 BLAKE2B newbeef SHA512 newcafe\n')
    mocker.patch('livecheck.special.checksum.catpkg_catpkgsplit',
                 return_value=('cat/foo', 'foo', 'r0', '1.0'))
    mocker.patch('livecheck.special.checksum.hash_url', return_value=('newbeef', 'newcafe', 4321))
    temp_file_mock = mocker.MagicMock()
    tf_mock = mocker.mock_open()
    temp_file_mock.open = tf_mock
    mocker.patch('livecheck.special.checksum.EbuildTempFile', return_value=temp_file_mock)
    temp_file_mock.__enter__.return_value = temp_file_mock
    temp_file_mock.__exit__.return_value = None
    manifest_open_mock = mocker.mock_open(read_data=manifest_content)
    mocker.patch('pathlib.Path.open', manifest_open_mock)
    update_checksum_metadata(ebuild, url, repo_root)
    handle = tf_mock()
    written_lines = ''.join(call.args[0] for call in handle.write.call_args_list)
    assert written_lines == expected_content



from __future__ import annotations

from typing import TYPE_CHECKING
import subprocess

from livecheck.utils.misc import check_program

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_check_program_success_no_version(mocker: MockerFixture) -> None:
    mock_run = mocker.patch('livecheck.utils.misc.sp.run')
    mock_run.return_value = mocker.Mock(stdout='Python 3.10.0\n')
    assert check_program('python', args=['--version']) is True
    mock_run.assert_called_once()


def test_check_program_success_with_version(mocker: MockerFixture) -> None:
    mock_run = mocker.patch('livecheck.utils.misc.sp.run')
    mock_run.return_value = mocker.Mock(stdout='Python 3.10.0\n')
    assert check_program('python', args=['--version'], min_version='3.9.0') is True


def test_check_program_version_too_low(mocker: MockerFixture) -> None:
    mock_run = mocker.patch('livecheck.utils.misc.sp.run')
    mock_run.return_value = mocker.Mock(stdout='Python 2.7.0\n')
    assert check_program('python', args=['--version'], min_version='3.0.0') is False


def test_check_program_version_not_found(mocker: MockerFixture) -> None:
    mock_run = mocker.patch('livecheck.utils.misc.sp.run')
    mock_run.return_value = mocker.Mock(stdout='no version info here\n')
    assert check_program('python', args=['--version'], min_version='3.0.0') is False


def test_check_program_run_raises_file_not_found(mocker: MockerFixture) -> None:
    mocker.patch('subprocess.run', side_effect=FileNotFoundError)
    assert check_program('not-real-cmd') is False


def test_check_program_run_raises_called_process_error(mocker: MockerFixture) -> None:
    mocker.patch('subprocess.run', side_effect=subprocess.CalledProcessError(1, 'cmd'))
    assert check_program('failing-cmd') is False


def test_check_program_version_value_error(mocker: MockerFixture) -> None:
    mock_run = mocker.patch('livecheck.utils.misc.sp.run')
    mocker.patch('livecheck.utils.misc.re.search', side_effect=ValueError)
    mock_run.return_value = mocker.Mock(stdout='Python version: unknown\n')
    assert check_program('python', args=['--version'], min_version='3.0.0') is False


def test_check_program_args_none(mocker: MockerFixture) -> None:
    mock_run = mocker.patch('livecheck.utils.misc.sp.run')
    mock_run.return_value = mocker.Mock(stdout='Python 3.10.0\n')
    assert check_program('python', args=None) is True



"""PyPI functions."""
from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import urlparse
import re

from livecheck.utils import get_content
from livecheck.utils.portage import get_last_version

from .utils import get_archive_extension

if TYPE_CHECKING:
    from collections.abc import Collection, Mapping

    from livecheck.settings import LivecheckSettings

__all__ = ('PYPI_METADATA', 'get_latest_pypi_metadata', 'get_latest_pypi_package', 'is_pypi')

PYPI_METADATA = 'pypi'

PYPI_DOWNLOAD_URL = 'https://pypi.org/pypi/%s/json'


def extract_project(url: str) -> str:
    parsed = urlparse(url)
    path = parsed.path.split('/')
    if (parsed.netloc in {'pypi', 'pypi.org', 'pypi.io', 'files.pythonhosted.org'}
            and len(path) > 3):  # noqa: PLR2004
        if path[2] == 'source':
            return path[4]
        if m := re.search(r'^(.*?)(?=-\d)', path[-1]):
            return m.group(1).strip()

    return ''


def get_url(ext: str, item: Collection[Mapping[str, str]]) -> str:
    for urls in item:
        if urls['url'].endswith(ext):
            return urls['url']
    for urls in item:
        if get_archive_extension(urls['url']):
            return str(urls['url'])
    return ''


def get_latest_pypi_package(src_uri: str, ebuild: str,
                            settings: LivecheckSettings) -> tuple[str, str]:
    """Get the latest version of a PyPI package."""
    project_name = extract_project(src_uri)
    return get_latest_pypi_package2(project_name, src_uri, ebuild, settings)


def get_latest_pypi_package2(project_name: str, src_uri: str, ebuild: str,
                             settings: LivecheckSettings) -> tuple[str, str]:
    url = PYPI_DOWNLOAD_URL % (project_name)
    ext = get_archive_extension(src_uri)

    results: list[dict[str, str]] = []
    if r := get_content(url):
        for release, item in r.json().get('releases', {}).items():
            results.extend([{'tag': release, 'url': get_url(ext, item)}])

        if last_version := get_last_version(results, '', ebuild, settings):
            return last_version['version'], last_version['url']

    return '', ''


def is_pypi(url: str) -> bool:
    """Check if the URL is a PyPI URL."""
    return bool(extract_project(url))


def get_latest_pypi_metadata(remote: str, ebuild: str,
                             settings: LivecheckSettings) -> tuple[str, str]:
    """Get the latest version of a PyPI package."""
    return get_latest_pypi_package2(remote, '', ebuild, settings)



"""Main command."""
from __future__ import annotations

from os import chdir
from pathlib import Path
from re import Match
from typing import TYPE_CHECKING
from urllib.parse import urlparse
import logging
import os
import re
import subprocess as sp

from defusedxml import ElementTree as ET  # noqa: N817
import click

from .constants import (
    SUBMODULES,
    TAG_NAME_FUNCTIONS,
)
from .settings import (
    TYPE_CHECKSUM,
    TYPE_COMMIT,
    TYPE_DAVINCI,
    TYPE_DIRECTORY,
    TYPE_METADATA,
    TYPE_NONE,
    TYPE_REGEX,
    TYPE_REPOLOGY,
    LivecheckSettings,
    gather_settings,
)
from .special.bitbucket import (
    BITBUCKET_METADATA,
    get_latest_bitbucket,
    get_latest_bitbucket_metadata,
    is_bitbucket,
)
from .special.checksum import get_latest_checksum_package, update_checksum_metadata
from .special.composer import (
    check_composer_requirements,
    remove_composer_url,
    update_composer_ebuild,
)
from .special.davinci import get_latest_davinci_package
from .special.directory import get_latest_directory_package
from .special.dotnet import check_dotnet_requirements, update_dotnet_ebuild
from .special.gist import get_latest_gist_package, is_gist
from .special.github import (
    GITHUB_METADATA,
    get_latest_github,
    get_latest_github_metadata,
    is_github,
)
from .special.gitlab import (
    GITLAB_METADATA,
    get_latest_gitlab,
    get_latest_gitlab_metadata,
    is_gitlab,
)
from .special.golang import update_go_ebuild
from .special.gomodule import (
    check_gomodule_requirements,
    remove_gomodule_url,
    update_gomodule_ebuild,
)
from .special.jetbrains import get_latest_jetbrains_package, is_jetbrains, update_jetbrains_ebuild
from .special.metacpan import (
    METACPAN_METADATA,
    get_latest_metacpan_metadata,
    get_latest_metacpan_package,
    is_metacpan,
)
from .special.nodejs import check_nodejs_requirements, remove_nodejs_url, update_nodejs_ebuild
from .special.package import get_latest_package, is_package
from .special.pecl import PECL_METADATA, get_latest_pecl_metadata, get_latest_pecl_package, is_pecl
from .special.pypi import PYPI_METADATA, get_latest_pypi_metadata, get_latest_pypi_package, is_pypi
from .special.regex import get_latest_regex_package
from .special.repology import get_latest_repology
from .special.rubygems import (
    RUBYGEMS_METADATA,
    get_latest_rubygems_metadata,
    get_latest_rubygems_package,
    is_rubygems,
)
from .special.sourceforge import (
    SOURCEFORGE_METADATA,
    get_latest_sourceforge_metadata,
    get_latest_sourceforge_package,
    is_sourceforge,
)
from .special.sourcehut import (
    SOURCEHUT_METADATA,
    get_latest_sourcehut,
    get_latest_sourcehut_metadata,
    is_sourcehut,
)
from .special.yarn import check_yarn_requirements, update_yarn_ebuild
from .utils import check_program, extract_sha, get_content, is_sha
from .utils.misc import setup_logging
from .utils.portage import (
    P,
    catpkg_catpkgsplit,
    catpkgsplit2,
    compare_versions,
    digest_ebuild,
    get_first_src_uri,
    get_highest_matches,
    get_repository_root_if_inside,
    remove_leading_zeros,
)

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    from .typing import PropTuple

log = logging.getLogger(__name__)

__all__ = ('main',)


def process_submodules(pkg_name: str, ref: str, contents: str, repo_uri: str) -> str:
    """Process submodules in the ebuild contents."""
    if pkg_name not in SUBMODULES:
        return contents
    offset_a, offset_b = ((1, 3) if 'api.github.com/repos/' in repo_uri else (0, 2))
    repo_root = '/'.join([x for x in urlparse(repo_uri).path.split('/') if x][offset_a:offset_b])
    ebuild_lines = contents.splitlines(keepends=True)
    for item in SUBMODULES[pkg_name]:
        name = item
        if isinstance(item, tuple):
            grep_for = f'{item[1]}="'
            name = item[0]
        else:
            grep_for = f'{Path(item).name.upper().replace("-", "_")}_SHA="'
        r = get_content(f'https://api.github.com/repos/{repo_root}/contents/{name}'
                        f'?ref={ref}')
        if not r.ok:
            continue
        remote_sha = r.json()['sha']
        for line in ebuild_lines:
            if (line.startswith(grep_for)
                    and (local_sha := line.split('=')[1].replace('"', '').strip()) != remote_sha):
                contents = contents.replace(local_sha, remote_sha)
    return contents


def log_unhandled_pkg(ebuild: str, src_uri: str) -> None:  # pragma: no cover
    """Log unhandled package name and ``SRC_URI`` at :py:ref:`logging.DEBUG` level."""
    log.debug('Unhandled: %s, SRC_URI: %s', ebuild, src_uri)


def parse_url(src_uri: str, ebuild: str, settings: LivecheckSettings, *,
              force_sha: bool) -> tuple[str, str, str, str]:
    """Parse a URL and return the last version, top hash, hash date, and URL."""
    parsed_uri = urlparse(src_uri)
    last_version = top_hash = hash_date = ''
    url = src_uri

    if not parsed_uri.hostname:
        return last_version, top_hash, hash_date, url

    log.debug('Parsed URI: %s', parsed_uri)
    if is_gist(src_uri):
        top_hash, hash_date = get_latest_gist_package(src_uri)
    elif is_github(src_uri):
        last_version, top_hash, hash_date = get_latest_github(src_uri,
                                                              ebuild,
                                                              settings,
                                                              force_sha=force_sha)
    elif is_sourcehut(src_uri):
        last_version, top_hash, hash_date = get_latest_sourcehut(src_uri,
                                                                 ebuild,
                                                                 settings,
                                                                 force_sha=force_sha)
    elif is_pypi(src_uri):
        last_version, url = get_latest_pypi_package(src_uri, ebuild, settings)
    elif is_jetbrains(src_uri):
        last_version = get_latest_jetbrains_package(ebuild, settings)
    elif is_gitlab(src_uri):
        last_version, top_hash, hash_date = get_latest_gitlab(src_uri,
                                                              ebuild,
                                                              settings,
                                                              force_sha=force_sha)
    elif is_package(src_uri):
        last_version = get_latest_package(src_uri, ebuild, settings)
    elif is_pecl(src_uri):
        last_version = get_latest_pecl_package(ebuild, settings)
    elif is_metacpan(src_uri):
        last_version = get_latest_metacpan_package(src_uri, ebuild, settings)
    elif is_rubygems(src_uri):
        last_version = get_latest_rubygems_package(ebuild, settings)
    elif is_sourceforge(src_uri):
        last_version = get_latest_sourceforge_package(src_uri, ebuild, settings)
    elif is_bitbucket(src_uri):
        last_version, top_hash, hash_date = get_latest_bitbucket(src_uri,
                                                                 ebuild,
                                                                 settings,
                                                                 force_sha=force_sha)
    else:
        log_unhandled_pkg(ebuild, src_uri)

    return last_version, top_hash, hash_date, url


def parse_metadata(repo_root: str, ebuild: str,
                   settings: LivecheckSettings) -> tuple[str, str, str, str]:
    """Parse ``metadata.xml`` for upstream information."""
    catpkg, _, _, _ = catpkg_catpkgsplit(ebuild)
    metadata_file = Path(repo_root) / catpkg / 'metadata.xml'
    if not metadata_file.exists():
        return '', '', '', ''
    try:
        root = ET.parse(metadata_file).getroot()
    except ET.ParseError:
        log.exception('Error parsing %s.', metadata_file)
        return '', '', '', ''
    for upstream in root.findall('upstream'):
        for subelem in upstream:
            last_version = top_hash = hash_date = url = ''
            if subelem.tag == 'remote-id':
                if not (remote := subelem.text.strip() if subelem.text else ''):
                    continue
                type_ = subelem.attrib['type']
                if GITHUB_METADATA in type_:
                    last_version, top_hash = get_latest_github_metadata(remote, ebuild, settings)
                if BITBUCKET_METADATA in type_:
                    last_version, top_hash = get_latest_bitbucket_metadata(remote, ebuild, settings)
                if GITLAB_METADATA in type_:
                    last_version, top_hash = get_latest_gitlab_metadata(
                        remote, type_, ebuild, settings)
                if SOURCEHUT_METADATA in type_:
                    last_version = get_latest_sourcehut_metadata(remote, ebuild, settings)
                if METACPAN_METADATA in type_:
                    last_version = get_latest_metacpan_metadata(remote, ebuild, settings)
                if PECL_METADATA in type_:
                    last_version = get_latest_pecl_metadata(remote, ebuild, settings)
                if RUBYGEMS_METADATA in type_:
                    last_version = get_latest_rubygems_metadata(remote, ebuild, settings)
                if SOURCEFORGE_METADATA in type_:
                    last_version = get_latest_sourceforge_metadata(remote, ebuild, settings)
                if PYPI_METADATA in type_:
                    last_version, url = get_latest_pypi_metadata(remote, ebuild, settings)
                if last_version or top_hash:
                    return last_version, top_hash, hash_date, url
    return '', '', '', ''


def extract_restrict_version(cp: str) -> tuple[str, str]:
    """Extract the restrict version from a package string."""
    if match := re.match(r'(.*?):(.*):-(.*)', cp):
        package, slot, version = match.groups()
        cleaned_string = f'{package}-{version}'
        return cleaned_string, slot
    return cp, ''


def get_props(search_dir: Path,
              repo_root: Path,
              settings: LivecheckSettings,
              names: Sequence[str] | None = None,
              exclude: Sequence[str] | None = None) -> Iterator[PropTuple]:
    """
    Get properties for packages in the search directory.

    Yields
    ------
    PropTuple

    Raises
    ------
    click.Abort
    """
    exclude = exclude or []
    if not names:
        names = [
            f'{path.parent.parent.name}/{path.parent.name}'
            for path in search_dir.glob('**/*.ebuild')
        ]
    matches_list = sorted(get_highest_matches(names, repo_root, settings))
    log.info('Found %d ebuild%s.', len(matches_list), 's' if len(matches_list) != 1 else '')
    if not matches_list:
        log.error('No matches!')
        raise click.Abort
    for _match in matches_list:
        match, settings.restrict_version_process = extract_restrict_version(_match)
        catpkg, cat, pkg, ebuild_version = catpkg_catpkgsplit(match)
        if catpkg in exclude or pkg in exclude:
            log.debug('Ignoring %s.', catpkg)
            continue
        src_uri = get_first_src_uri(match, repo_root)
        if cat.startswith(('acct-', 'virtual')) or settings.type_packages.get(catpkg) == TYPE_NONE:
            log.debug('Ignoring %s.', catpkg)
            continue
        log.info('Processing: %s | Version: %s', catpkg, ebuild_version)
        last_version = hash_date = top_hash = url = ''
        ebuild = Path(repo_root) / catpkg / f'{pkg}-{ebuild_version}.ebuild'
        egit, branch = get_egit_repo(ebuild)
        if egit:
            old_sha = get_old_sha(ebuild, '')
            egit = egit + '/commit/' + old_sha
        if branch:
            settings.branches[catpkg] = branch
        if catpkg in settings.sync_version:
            matches_sync = get_highest_matches([settings.sync_version[catpkg]], None, settings)
            if not matches_sync:
                log.error('No matches for %s.', catpkg)
                continue
            _, _, _, last_version = catpkg_catpkgsplit(matches_sync[0])
            # remove -r* from version
            last_version = re.sub(r'-r\d+$', '', last_version)
        if settings.type_packages.get(catpkg) == TYPE_DAVINCI:
            last_version = get_latest_davinci_package(pkg)
        elif settings.type_packages.get(catpkg) == TYPE_METADATA:
            last_version, top_hash, hash_date, url = parse_metadata(str(repo_root), match, settings)
        elif settings.type_packages.get(catpkg) == TYPE_DIRECTORY:
            url, _ = settings.custom_livechecks[catpkg]
            last_version, url = get_latest_directory_package(url, match, settings)
        elif settings.type_packages.get(catpkg) == TYPE_REPOLOGY:
            package, _ = settings.custom_livechecks[catpkg]
            last_version = get_latest_repology(match, settings, package)
        elif settings.type_packages.get(catpkg) == TYPE_REGEX:
            url, regex = settings.custom_livechecks[catpkg]
            last_version, hash_date, url = get_latest_regex_package(match, url, regex, settings)
        elif settings.type_packages.get(catpkg) == TYPE_CHECKSUM:
            last_version, hash_date, url = get_latest_checksum_package(
                src_uri, match, str(repo_root))
        elif settings.type_packages.get(catpkg) == TYPE_COMMIT:
            last_version, top_hash, hash_date, url = parse_url(egit,
                                                               match,
                                                               settings,
                                                               force_sha=True)
        else:
            if egit:
                last_version, top_hash, hash_date, url = parse_url(egit,
                                                                   match,
                                                                   settings,
                                                                   force_sha=True)
            if not last_version and not top_hash:
                last_version, top_hash, hash_date, url = parse_url(src_uri,
                                                                   match,
                                                                   settings,
                                                                   force_sha=False)
            if not last_version and not top_hash:
                last_version, top_hash, hash_date, url = parse_metadata(
                    str(repo_root), match, settings)
            # Try check for homepage
            homes = [
                x
                for x in ' '.join(P.aux_get(match, ['HOMEPAGE'], mytree=str(repo_root))).split(' ')
                if x
            ]
            for home in homes:
                if not last_version and not top_hash:
                    last_version, top_hash, hash_date, url = parse_url(home,
                                                                       match,
                                                                       settings,
                                                                       force_sha=False)
            if not last_version and not top_hash:
                last_version = get_latest_repology(match, settings)
            # Only check directory if no other method was found
            if not last_version and not top_hash:
                last_version, url = get_latest_directory_package(src_uri, match, settings)
                for home in homes:
                    last_version, url = get_latest_directory_package(home, match, settings)
                    if last_version:
                        break

        if last_version or top_hash:
            log.debug('Inserting %s: %s -> %s : %s', catpkg, ebuild_version, last_version, top_hash)
            yield (cat, pkg, ebuild_version, last_version, top_hash, hash_date, url)
        else:
            log.debug('Ignoring %s. Update not available.', catpkg)


def get_old_sha(ebuild: Path, url: str) -> str:
    sha_pattern = re.compile(r'(SHA|COMMIT|EGIT_COMMIT)=["\']?([a-f0-9]{40})')

    with Path(ebuild).open(encoding='utf-8') as file:
        for line in file:
            if match := sha_pattern.search(line):
                return match.group(2)

    last_part = urlparse(url).path.rsplit('/', 1)[-1] if '/' in url else url
    return extract_sha(last_part)


def get_egit_repo(ebuild: Path) -> tuple[str, str]:
    egit = branch = ''
    with Path(ebuild).open(encoding='utf-8') as file:
        for line in file:
            if match := re.compile(r'^EGIT_REPO_URI=(["\'])?(.*)\1').search(line):
                egit = match.group(2)
            if match := re.compile(r'^EGIT_BRANCH=(["\'])?(.*)\1').search(line):
                branch = match.group(2)
    return egit, branch


def str_version(version: str, sha: str) -> str:
    return version + f' ({sha})' if sha else version


DATE_LENGTH_8 = 8
DATE_LENGTH_6 = 6


def replace_date_in_ebuild(ebuild: str, new_date: str, cp: str) -> str:
    short_date = new_date[2:]
    pattern = re.compile(r'(\d{4,8})')

    def replace_match(match: Match[str]) -> str:
        matched_text = match.group(0)
        if len(matched_text) == DATE_LENGTH_8 and matched_text.isdigit():
            return new_date
        if len(matched_text) == DATE_LENGTH_6 and matched_text.isdigit():
            return short_date
        return matched_text

    n = pattern.sub(replace_match, ebuild)

    _, _, old_version, _ = catpkgsplit2(f'{cp}-{ebuild}')
    _, _, new_version, _ = catpkgsplit2(f'{cp}-{n}')
    return str(new_version) if old_version != new_version else n


def execute_hooks(hook_dir: Path | None, action: str, search_dir: Path, cp: str,
                  str_old_version: str, str_new_version: str, old_sha: str, new_sha: str,
                  hash_date: str) -> None:
    if not hook_dir:
        return
    hook_path = hook_dir / action
    if not hook_path.is_dir():
        return
    for hook in sorted(hook_path.iterdir()):
        if hook.is_file() and os.access(hook, os.X_OK):
            log.debug('Running hook {hook}')
            result = sp.run([
                hook, search_dir, cp, str_old_version, str_new_version, old_sha, new_sha, hash_date
            ],
                            check=False)
            if result.returncode != 0:
                click.echo(f'Error running hook {hook}.', err=True)
                raise click.Abort


def do_main(*, cat: str, ebuild_version: str, pkg: str, search_dir: Path,
            settings: LivecheckSettings, last_version: str, top_hash: str, hash_date: str, url: str,
            hook_dir: Path | None) -> None:
    cp = f'{cat}/{pkg}'
    ebuild = Path(search_dir) / cp / f'{pkg}-{ebuild_version}.ebuild'
    old_sha = ''
    if update_sha_too_source := settings.sha_sources.get(cp, None):
        log.debug('Package also needs a SHA update.')
        _, top_hash, hash_date, _ = parse_url(update_sha_too_source,
                                              f'{cp}-{ebuild_version}',
                                              settings,
                                              force_sha=True)

        # if empty, it means that the source is not supported
        if not top_hash:
            log.warning('Could not get new SHA for %s.', update_sha_too_source)
            return
    if top_hash:
        old_sha = get_old_sha(ebuild, url)
        if len(old_sha) == 7:  # noqa: PLR2004
            top_hash = top_hash[:7]
        log.debug('Get old_sha = %s', old_sha)
    if not last_version:
        last_version = ebuild_version
    if hash_date:
        last_version = replace_date_in_ebuild(last_version, hash_date, cp)
    if last_version == ebuild_version and old_sha != top_hash and old_sha and top_hash:
        _, _, new_version, new_revision = catpkgsplit2(f'{cp}-{last_version}')
        new_revision = 'r' + str(int(new_revision[1:]) + 1)
        log.debug('Incrementing revision to %s.', new_revision)
        last_version = f'{new_version}-{new_revision}'
    log.debug('top_hash = %s', last_version)

    # Remove leading zeros to prevent issues with version comparison
    last_version = remove_leading_zeros(last_version)
    ebuild_version = remove_leading_zeros(ebuild_version)

    log.debug('Comparing current ebuild version %s with live version %s.', ebuild_version,
              last_version)
    if compare_versions(ebuild_version, last_version):
        dn = Path(ebuild).parent
        new_filename = f'{dn}/{pkg}-{last_version}.ebuild'
        _, _, _, new_version = catpkg_catpkgsplit(f'{cp}-{last_version}')
        _, _, _, old_version = catpkg_catpkgsplit(f'{cp}-{ebuild_version}')
        log.debug('Migrating from %s to %s.', ebuild, new_filename)
        no_auto_update_str = ' (no_auto_update)' if cp in settings.no_auto_update else ''
        str_new_version = str_version(new_version, top_hash)
        str_old_version = str_version(old_version, old_sha)
        log.info('%s/%s: %s -> %s%s', cat, pkg, str_old_version, str_new_version,
                 no_auto_update_str)

        if settings.auto_update_flag and cp not in settings.no_auto_update:
            # First check requirements before update
            if ((  # noqa: PLR0916
                    cp in settings.dotnet_projects and not check_dotnet_requirements())
                    or (cp in settings.composer_packages and not check_composer_requirements())
                    or (cp in settings.yarn_base_packages and not check_yarn_requirements())
                    or (cp in settings.nodejs_packages and not check_nodejs_requirements())
                    or (cp in settings.gomodule_packages and not check_gomodule_requirements())):
                log.warning('Update is not possible.')
                return
            old_content = content = Path(ebuild).read_text(encoding='utf-8')
            # Only update the version if it is not a commit
            if top_hash and old_sha:
                content = content.replace(old_sha, top_hash)
            ps_ref = top_hash
            if not is_sha(top_hash) and cp in TAG_NAME_FUNCTIONS:
                ps_ref = TAG_NAME_FUNCTIONS[cp](top_hash)
            content = process_submodules(cp, ps_ref, content, url)
            dn = Path(ebuild).parent
            log.debug('%s -> %s', ebuild, new_filename)
            if settings.keep_old.get(cp, not settings.keep_old_flag):
                try:
                    if settings.git_flag:
                        sp.run(('git', 'mv', str(ebuild), new_filename), check=True)
                    else:
                        ebuild.rename(new_filename)
                except (sp.CalledProcessError, OSError):
                    log.exception('Error moving `%s` to `%s`.', ebuild, new_filename)
                    return

            try:
                Path(new_filename).write_text(content, encoding='utf-8')
            except OSError:
                log.exception('Error writing `%s`.', new_filename)
                return
            execute_hooks(hook_dir, 'pre', search_dir, cp, ebuild_version, last_version, old_sha,
                          top_hash, hash_date)
            # We do not check the digest because it may happen that additional files need to be
            # created that have not yet been generated until the main ebuild is downloaded for
            # the first time and the hooks create those additional files.
            # And you cannot remove the digest generation either, but rather the repositories
            # that do not have thin-Manifests ( metadata/layout.conf -> thin-manifests = true)
            # see: https://devmanual.gentoo.org/general-concepts/manifest/index.html
            digest_ebuild(new_filename)
            fetchlist = P.getFetchMap(f'{cp}-{last_version}')
            # Stores the content so that it can be recovered because it had to be modified
            old_content = content
            # First pass
            # Remove URLs that do not yet exist to be able to correctly generate the digest
            if cp in settings.gomodule_packages:
                content = remove_gomodule_url(content)
            if cp in settings.nodejs_packages:
                content = remove_nodejs_url(content)
            if cp in settings.composer_packages:
                content = remove_composer_url(content)
            if old_content != content:
                Path(new_filename).write_text(content, encoding='utf-8')
            if not digest_ebuild(new_filename):
                log.error('Error digesting %s.', new_filename)
                return
            # Second pass
            # Update ebuild or download news file
            if cp in settings.yarn_base_packages:
                update_yarn_ebuild(new_filename, settings.yarn_base_packages[cp], pkg,
                                   settings.yarn_packages.get(cp))
            if settings.type_packages.get(cp) == TYPE_CHECKSUM:
                update_checksum_metadata(f'{cp}-{last_version}', url, str(search_dir))
            if cp in settings.go_sum_uri:
                update_go_ebuild(new_filename, top_hash, settings.go_sum_uri[cp])
            if cp in settings.dotnet_projects:
                update_dotnet_ebuild(new_filename, settings.dotnet_projects[cp])
            if cp in settings.jetbrains_packages:
                update_jetbrains_ebuild(new_filename)
            if cp in settings.nodejs_packages:
                update_nodejs_ebuild(new_filename, settings.nodejs_path[cp], fetchlist)
            if cp in settings.gomodule_packages:
                update_gomodule_ebuild(new_filename, settings.gomodule_path[cp], fetchlist)
            if cp in settings.composer_packages:
                update_composer_ebuild(new_filename, settings.composer_path[cp], fetchlist)
            # Restore original ebuild content
            if old_content != content:
                Path(new_filename).write_text(old_content, encoding='utf-8')
                if not digest_ebuild(new_filename):
                    log.error('Error digesting %s.', new_filename)
                    return
            if settings.git_flag and sp.run(
                ('ebuild', new_filename, 'digest'), check=False).returncode == 0:
                sp.run(('git', 'add', new_filename, str(Path(search_dir) / cp / 'Manifest')),
                       check=True)
                try:
                    sp.run(('pkgdev', 'commit'), cwd=Path(search_dir) / cp, check=True)
                except sp.CalledProcessError:
                    log.exception('Error committing %s.', new_filename)
            execute_hooks(hook_dir, 'post', search_dir, cp, ebuild_version, last_version, old_sha,
                          top_hash, hash_date)


@click.command(context_settings={'help_option_names': ['-h', '--help']})
@click.option('-a', '--auto-update', is_flag=True, help='Rename and modify ebuilds.')
@click.option('-d', '--debug', is_flag=True, help='Enable debug logging.')
@click.option('-D', '--development', is_flag=True, help='Include development packages.')
@click.option('-e', '--exclude', multiple=True, help='Exclude package(s) from updates.')
@click.option('-g', '--git', is_flag=True, help='Use git and pkgdev to make changes.')
@click.option('-H',
              '--hook-dir',
              default=None,
              help='Run a hook directory scripts with various parameters.',
              type=click.Path(file_okay=False, exists=True, resolve_path=True, path_type=Path))
@click.option('-k', '--keep-old', is_flag=True, help='Keep old ebuild versions.')
@click.option('-p', '--progress', is_flag=True, help='Enable progress logging.')
@click.option('-W',
              '--working-dir',
              default='.',
              help='Working directory. Should be a port tree root.',
              type=click.Path(file_okay=False,
                              exists=True,
                              resolve_path=True,
                              readable=True,
                              path_type=Path))
@click.argument('package_names', nargs=-1)
def main(working_dir: Path,
         exclude: tuple[str, ...] | None = None,
         hook_dir: Path | None = None,
         package_names: tuple[str, ...] | list[str] | None = None,
         *,
         auto_update: bool = False,
         debug: bool = False,
         development: bool = False,
         git: bool = False,
         keep_old: bool = False,
         progress: bool = False) -> None:
    """Update ebuilds to their latest versions."""  # noqa: DOC501
    setup_logging(debug=debug)
    chdir(working_dir)
    if exclude:
        log.debug('Excluding %s.', ', '.join(exclude))
    search_dir = working_dir or Path()
    if auto_update and not os.access(search_dir, os.W_OK):
        msg = f'The directory "{working_dir}" must be writable because --auto-update is enabled.'
        raise click.ClickException(msg)
    repo_root, repo_name = get_repository_root_if_inside(search_dir)
    if not repo_root:
        log.error('Not inside a repository configured in repos.conf.')
        raise click.Abort
    if git:
        if not auto_update:
            log.error('Git option requires --auto-update.')
            raise click.Abort
        if not Path(repo_root, '.git').is_dir():
            log.error('Directory %s is not a git repository.', repo_root)
            raise click.Abort
        # Check if .git is a writeable directory
        if not os.access(Path(repo_root, '.git'), os.W_OK):
            log.error('Directory %s/.git is not writable.', repo_root)
            raise click.Abort
        # Check if git is installed
        if not check_program('git', ['--version']):
            log.error('Git is not installed.')
            raise click.Abort
        # Check if pkgdev is installed
        if not check_program('pkgdev', ['--version']):
            log.error('pkgdev is not installed.')
            raise click.Abort
    log.debug('search_dir=%s repo_root=%s repo_name=%s', search_dir, repo_root, repo_name)
    settings = gather_settings(Path(repo_root))

    # update flags in settings
    settings.auto_update_flag = auto_update
    settings.debug_flag = debug
    settings.development_flag = development
    settings.git_flag = git
    settings.keep_old_flag = keep_old
    settings.progress_flag = progress

    package_names = sorted(package_names or [])
    cat = pkg = None
    try:
        for cat, pkg, ebuild_version, last_version, top_hash, hash_date, url in get_props(
                search_dir, Path(repo_root), settings, package_names, exclude):
            do_main(cat=cat,
                    pkg=pkg,
                    last_version=last_version,
                    top_hash=top_hash,
                    hash_date=hash_date,
                    url=url,
                    search_dir=Path(repo_root),
                    settings=settings,
                    ebuild_version=ebuild_version,
                    hook_dir=hook_dir)
    except Exception:
        if cat and pkg:
            log.exception('Exception while checking %s/%s.', cat, pkg)
        raise



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from livecheck.special import repology
import pytest

if TYPE_CHECKING:
    from unittest.mock import Mock

    from pytest_mock import MockerFixture


@pytest.fixture
def mock_settings(mocker: MockerFixture) -> Any:
    settings = mocker.Mock()
    settings.is_devel.return_value = False
    return settings


@pytest.fixture
def mock_catpkg_catpkgsplit(mocker: MockerFixture) -> Any:
    return mocker.patch('livecheck.special.repology.catpkg_catpkgsplit',
                        return_value=('cat/pkg', None, 'pkg', None))


@pytest.fixture
def mock_get_last_version(mocker: MockerFixture) -> Any:
    return mocker.patch('livecheck.special.repology.get_last_version')


def test_get_latest_repology_success(mock_catpkg_catpkgsplit: Mock, mocker: MockerFixture,
                                     mock_get_last_version: Mock, mock_settings: Mock) -> None:
    # Mock the response from get_content
    mock_response = mocker.Mock()
    mock_response.json.return_value = [
        {
            'srcname': 'pkg',
            'status': 'stable',
            'version': '1.2.3'
        },
        {
            'srcname': 'pkg',
            'status': 'devel',
            'version': '2.0.0'
        },
        {
            'srcname': 'other',
            'status': 'stable',
            'version': '9.9.9'
        },
    ]
    mock_get_content = mocker.patch('livecheck.special.repology.get_content')
    mock_get_content.return_value = mock_response
    mock_get_last_version.return_value = {'version': '1.2.3'}

    result = repology.get_latest_repology('cat/pkg', mock_settings)
    assert result == '1.2.3'
    mock_get_content.assert_called_once()
    mock_get_last_version.assert_called_once()


def test_get_latest_repology_with_package_param(mock_catpkg_catpkgsplit: Mock,
                                                mocker: MockerFixture, mock_get_last_version: Mock,
                                                mock_settings: Mock) -> None:
    mock_response = mocker.Mock()
    mock_response.json.return_value = [
        {
            'srcname': 'custom-pkg',
            'status': 'stable',
            'version': '4.5.6'
        },
    ]
    mock_get_content = mocker.patch('livecheck.special.repology.get_content')
    mock_get_content.return_value = mock_response
    mock_get_last_version.return_value = {'version': '4.5.6'}

    result = repology.get_latest_repology('cat/pkg', mock_settings, package='custom-pkg')
    assert result == '4.5.6'
    mock_get_content.assert_called_once()
    mock_get_last_version.assert_called_once()


def test_get_latest_repology_fallback_url(mock_catpkg_catpkgsplit: Mock, mocker: MockerFixture,
                                          mock_get_last_version: Mock, mock_settings: Mock) -> None:
    # First call returns None, second call returns a valid response
    mock_response = mocker.Mock()
    mock_response.json.return_value = [
        {
            'srcname': 'pkg',
            'status': 'stable',
            'version': '7.8.9'
        },
    ]
    mock_get_content = mocker.patch('livecheck.special.repology.get_content')
    mock_get_content.side_effect = [None, mock_response]
    mock_get_last_version.return_value = {'version': '7.8.9'}

    result = repology.get_latest_repology('cat/pkg', mock_settings)
    assert result == '7.8.9'
    assert mock_get_content.call_count == 2
    mock_get_last_version.assert_called_once()


def test_get_latest_repology_no_content(mock_catpkg_catpkgsplit: Mock, mocker: MockerFixture,
                                        mock_settings: Mock) -> None:
    mock_get_content = mocker.patch('livecheck.special.repology.get_content')
    mock_get_content.side_effect = [None, None]
    result = repology.get_latest_repology('cat/pkg', mock_settings)
    assert not result


def test_get_latest_repology_no_matching_release(mock_catpkg_catpkgsplit: Mock,
                                                 mocker: MockerFixture, mock_get_last_version: Mock,
                                                 mock_settings: Mock) -> None:
    mock_response = mocker.Mock()
    mock_response.json.return_value = [
        {
            'srcname': 'other',
            'status': 'stable',
            'version': '0.0.1'
        },
    ]
    mock_get_content = mocker.patch('livecheck.special.repology.get_content')
    mock_get_content.return_value = mock_response
    mock_get_last_version.return_value = None

    result = repology.get_latest_repology('cat/pkg', mock_settings)
    assert not result


def test_get_latest_repology_devel_status_allowed(mock_catpkg_catpkgsplit: Mock,
                                                  mocker: MockerFixture,
                                                  mock_get_last_version: Mock,
                                                  mock_settings: Mock) -> None:
    mock_settings.is_devel.return_value = True
    mock_response = mocker.Mock()
    mock_response.json.return_value = [
        {
            'srcname': 'pkg',
            'status': 'devel',
            'version': '3.3.3'
        },
    ]
    mock_get_content = mocker.patch('livecheck.special.repology.get_content')
    mock_get_content.return_value = mock_response
    mock_get_last_version.return_value = {'version': '3.3.3'}

    result = repology.get_latest_repology('cat/pkg', mock_settings)
    assert result == '3.3.3'



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from livecheck.special.gitlab import (
    extract_domain_and_namespace,
    get_latest_gitlab,
    get_latest_gitlab_metadata,
    is_gitlab,
)
import pytest

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

test_cases = {
    'valid_gitlab_url': {
        'url': 'https://gitlab.com/group/project',
        'expected': ('gitlab.com', 'group/project', 'project'),
        'is_gitlab': True
    },
    'invalid_gitlab_url': {
        'url': 'https://notgitlab.com/group/project',
        'expected': ('', '', ''),
        'is_gitlab': False
    },
    'invalid_gitlab_url2': {
        'url': 'https://gitlab.com',
        'expected': ('', '', ''),
        'is_gitlab': False
    },
    'invalid_gitlab_url3': {
        'url': 'https://gitlab.com/group/',
        'expected': ('', '', ''),
        'is_gitlab': False
    },
    'gitlab_es_url': {
        'url': 'https://gitlab.es/group/project',
        'expected': ('', '', ''),
        'is_gitlab': False
    },
    'example_gitlab_url': {
        'url': 'https://example.gitlab.com/group/project',
        'expected': ('', '', ''),
        'is_gitlab': False
    },
    'gitlab_example_url': {
        'url': 'https://gitlab.example.com/group/project',
        'expected': ('gitlab.example.com', 'group/project', 'project'),
        'is_gitlab': True
    },
    'example_com_url': {
        'url': 'https://example.com/group/project',
        'expected': ('', '', ''),
        'is_gitlab': False
    },
    'gitlab_merge_request_url': {
        'url': 'https://gitlab.com/group/project/-/merge_requests',
        'expected': ('gitlab.com', 'group/project', 'project'),
        'is_gitlab': True
    },
    'gitlab_subgroup_url': {
        'url': 'https://gitlab.com/group/subgroup/project',
        'expected': ('gitlab.com', 'group/subgroup/project', 'project'),
        'is_gitlab': True
    },
    'sys-apps/udev-usb-sync': {
        'url': 'https://gitlab.manjaro.org/fhdk/udev-usb-sync',
        'expected': ('gitlab.manjaro.org', 'fhdk/udev-usb-sync', 'udev-usb-sync'),
        'is_gitlab': True
    },
    'x11-misc/xdg-utils': {
        'url':
            'https://gitlab.freedesktop.org/xdg/xdg-utils/-/archive/v1.2.1/xdg-utils-1.2.1.tar.bz2',
        'expected': ('gitlab.freedesktop.org', 'xdg/xdg-utils', 'xdg-utils'),
        'is_gitlab':
            True
    },
}


@pytest.mark.parametrize('test_case', test_cases.values(), ids=test_cases.keys())
def test_extract_domain_and_namespace(test_case: dict[str, Any]) -> None:
    assert extract_domain_and_namespace(test_case['url']) == test_case['expected']


@pytest.mark.parametrize('test_case', test_cases.values(), ids=test_cases.keys())
def test_is_gitlab(test_case: dict[str, Any]) -> None:
    assert is_gitlab(test_case['url']) == test_case['is_gitlab']


@pytest.mark.parametrize(('url', 'ebuild', 'force_sha', 'content_return', 'expected_version',
                          'expected_hash', 'expected_date'), [
                              (
                                  'https://gitlab.com/group/project',
                                  'project-1.0.ebuild',
                                  False,
                                  [{
                                      'name': 'v2.0',
                                      'commit': {
                                          'id': 'abc123'
                                      }
                                  }],
                                  '2.0',
                                  '',
                                  '',
                              ),
                              (
                                  'https://gitlab.com/group/project',
                                  'project-1.0.ebuild',
                                  True,
                                  [{
                                      'name': 'v2.0',
                                      'commit': {
                                          'id': 'abc123'
                                      }
                                  }],
                                  '2.0',
                                  'abc123',
                                  '',
                              ),
                              (
                                  'https://gitlab.com/group/project',
                                  'project-1.0.ebuild',
                                  False,
                                  [],
                                  '',
                                  '',
                                  '',
                              ),
                          ])
def test_get_latest_gitlab(
        mocker: MockerFixture,
        url: str,
        ebuild: str,
        force_sha: bool,  # noqa: FBT001
        content_return: str,
        expected_version: str,
        expected_hash: str,
        expected_date: str) -> None:
    mock_content = mocker.Mock()
    mock_content.json.return_value = content_return
    mocker.patch('livecheck.special.gitlab.get_content', return_value=mock_content)
    if content_return:
        mocker.patch('livecheck.special.gitlab.get_last_version',
                     return_value={
                         'version': '2.0',
                         'id': 'abc123'
                     })
    else:
        mocker.patch('livecheck.special.gitlab.get_last_version', return_value=None)
    mocker.patch('livecheck.utils.is_sha', return_value=False)
    log_unhandled_commit = mocker.patch('livecheck.special.gitlab.log_unhandled_commit')
    result = get_latest_gitlab(url, ebuild, mocker.Mock(), force_sha=force_sha)
    assert result == (expected_version, expected_hash, expected_date)
    assert not log_unhandled_commit.called


def test_get_latest_gitlab_package_no_content(mocker: MockerFixture) -> None:
    url = 'https://gitlab.com/group/project'
    ebuild = 'project-1.0.ebuild'
    mocker.patch('livecheck.special.gitlab.get_content', return_value=None)
    result = get_latest_gitlab(url, ebuild, mocker.Mock(), force_sha=False)
    assert result == ('', '', '')


def test_get_latest_gitlab_with_sha(mocker: MockerFixture) -> None:
    url = 'https://gitlab.com/group/project/commit/abc123'
    ebuild = 'project-1.0.ebuild'
    mocker.patch('livecheck.special.gitlab.is_sha', return_value=True)
    log_unhandled_commit = mocker.patch('livecheck.special.gitlab.log_unhandled_commit')
    result = get_latest_gitlab(url, ebuild, mocker.Mock(), force_sha=False)
    assert result == ('', '', '')
    log_unhandled_commit.assert_called_once_with(ebuild, url)


@pytest.mark.parametrize(
    ('remote', '_type', 'ebuild', 'package_return', 'expected'),
    [
        (
            'group/project',
            'gitlab',
            'project-1.0.ebuild',
            ('2.0', 'abc123'),
            ('2.0', 'abc123'),
        ),
        (
            'fhdk/udev-usb-sync',
            'manjaro-gitlab',
            'udev-usb-sync-1.0.ebuild',
            ('1.5', 'def456'),
            ('1.5', 'def456'),
        ),
        (
            'xdg/xdg-utils',
            'freedesktop-gitlab',
            'xdg-utils-1.2.1.ebuild',
            ('', ''),
            ('', ''),
        ),
    ],
)
def test_get_latest_gitlab_metadata(mocker: MockerFixture, remote: str, _type: str, ebuild: str,
                                    package_return: tuple[str, str], expected: tuple[str,
                                                                                     str]) -> None:
    mocker.patch(
        'livecheck.special.gitlab.get_latest_gitlab_package',
        return_value=package_return,
    )
    result = get_latest_gitlab_metadata(remote, _type, ebuild, mocker.Mock())
    assert result == expected



# ruff: noqa: S108
from __future__ import annotations

from typing import TYPE_CHECKING
import subprocess as sp

from livecheck.special.gomodule import remove_gomodule_url, update_gomodule_ebuild

if TYPE_CHECKING:
    from pytest_mock import MockerFixture
from livecheck.special import gomodule


def test_remove_gomodule_url_removes_vendor_line(mocker: MockerFixture) -> None:
    mock_remove = mocker.patch('livecheck.special.gomodule.remove_url_ebuild',
                               return_value='filtered content')
    ebuild_content = 'SRC_URI="https://example.com/foo-vendor.tar.xz"\n'
    result = remove_gomodule_url(ebuild_content)
    mock_remove.assert_called_once_with(ebuild_content, '-vendor.tar.xz')
    assert result == 'filtered content'


def test_remove_gomodule_url_no_vendor_line(mocker: MockerFixture) -> None:
    mock_remove = mocker.patch('livecheck.special.gomodule.remove_url_ebuild',
                               side_effect=lambda content, _: content)
    ebuild_content = 'SRC_URI="https://example.com/foo.tar.gz"\n'
    result = remove_gomodule_url(ebuild_content)
    mock_remove.assert_called_once_with(ebuild_content, '-vendor.tar.xz')
    assert result == ebuild_content


def test_update_gomodule_ebuild_success(mocker: MockerFixture) -> None:
    mock_search = mocker.patch('livecheck.special.gomodule.search_ebuild',
                               return_value=('/some/path', '/tmp/dir'))
    mock_run = mocker.patch('livecheck.special.gomodule.sp.run', return_value=None)
    mock_build = mocker.patch('livecheck.special.gomodule.build_compress')
    ebuild = 'dummy.ebuild'
    path = '/some/path'
    fetchlist = {'foo': ('bar',)}
    update_gomodule_ebuild(ebuild, path, fetchlist)
    mock_search.assert_called_once_with(ebuild, 'go.mod', path)
    mock_run.assert_called_once_with(('go', 'mod', 'vendor'), cwd='/some/path', check=True)
    mock_build.assert_called_once_with('/tmp/dir', '/some/path', 'vendor', '-vendor.tar.xz',
                                       fetchlist)


def test_update_gomodule_ebuild_no_go_mod_path(mocker: MockerFixture) -> None:
    mock_search = mocker.patch('livecheck.special.gomodule.search_ebuild',
                               return_value=(None, None))
    mock_run = mocker.patch('livecheck.special.gomodule.sp.run')
    mock_build = mocker.patch('livecheck.special.gomodule.build_compress')
    update_gomodule_ebuild('ebuild', None, {})
    mock_search.assert_called_once()
    mock_run.assert_not_called()
    mock_build.assert_not_called()


def test_update_gomodule_ebuild_subprocess_error(mocker: MockerFixture) -> None:
    mock_search = mocker.patch('livecheck.special.gomodule.search_ebuild',
                               return_value=('/some/path', '/tmp/dir'))
    mock_run = mocker.patch('livecheck.special.gomodule.sp.run',
                            side_effect=sp.CalledProcessError(1, 'go mod vendor'))
    mock_build = mocker.patch('livecheck.special.gomodule.build_compress')
    update_gomodule_ebuild('ebuild', '/some/path', {})
    mock_search.assert_called_once()
    mock_run.assert_called_once()
    mock_build.assert_not_called()


def test_check_gomodule_requirements_success(mocker: MockerFixture) -> None:
    mock_check_program = mocker.patch('livecheck.special.gomodule.check_program', return_value=True)
    result = gomodule.check_gomodule_requirements()
    mock_check_program.assert_called_once_with('go', ['version'])
    assert result is True


def test_check_gomodule_requirements_failure(mocker: MockerFixture) -> None:
    mock_check_program = mocker.patch('livecheck.special.gomodule.check_program',
                                      return_value=False)
    mock_logger = mocker.patch('livecheck.special.gomodule.logger')
    result = gomodule.check_gomodule_requirements()
    mock_check_program.assert_called_once_with('go', ['version'])
    mock_logger.error.assert_called_once_with('go is not installed')
    assert result is False



"""MetaCPAN functions."""
from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import urlparse
import re

from livecheck.utils import get_content
from livecheck.utils.portage import get_last_version

if TYPE_CHECKING:
    from livecheck.settings import LivecheckSettings

__all__ = ('METACPAN_METADATA', 'get_latest_metacpan_metadata', 'get_latest_metacpan_package',
           'is_metacpan')

METACPAN_METADATA = 'cpan'
METACPAN_DOWNLOAD_URL1 = 'https://fastapi.metacpan.org/v1/release/_search?q=distribution:%s'
METACPAN_DOWNLOAD_URL2 = 'https://fastapi.metacpan.org/v1/release/%s'


def extract_perl_package(url: str) -> str:
    parsed = urlparse(url)
    if parsed.netloc not in {'metacpan.org', 'cpan'}:
        return ''

    match = re.search(r'/([^/]+)-[\d.]+\.*', parsed.path)
    return match.group(1) if match else ''


def get_latest_metacpan_package(url: str, ebuild: str, settings: LivecheckSettings) -> str:
    """Get the latest version of a MetaCPAN package."""
    package_name = extract_perl_package(url)
    return get_latest_metacpan_package2(package_name, ebuild, settings)


def get_latest_metacpan_package2(package_name: str, ebuild: str,
                                 settings: LivecheckSettings) -> str:
    results: list[dict[str, str]] = []
    url = METACPAN_DOWNLOAD_URL1 % (package_name)
    if r := get_content(url):
        for hit in r.json().get('hits', {}).get('hits', []):
            results.extend([{'tag': hit['_source']['version']}])

    # Many times it does not exist as in the previous list,
    # that is why the latest version is checked again.
    url = METACPAN_DOWNLOAD_URL2 % (package_name)
    if r := get_content(url):
        results.append({'tag': r.json().get('version')})

    last_version = get_last_version(results, package_name, ebuild, settings)
    if last_version:
        return last_version['version']

    return ''


def is_metacpan(url: str) -> bool:
    """Check if the URL is a MetaCPAN URL."""
    return bool(extract_perl_package(url))


def get_latest_metacpan_metadata(remote: str, ebuild: str, settings: LivecheckSettings) -> str:
    """Get the latest version of a MetaCPAN package."""
    return get_latest_metacpan_package2(remote, ebuild, settings)



from __future__ import annotations

from typing import IO, TYPE_CHECKING, Any

from livecheck.special.dotnet import (
    NoNugetsEnding,
    NoNugetsFound,
    check_dotnet_requirements,
    dotnet_restore,
    update_dotnet_ebuild,
)
from typing_extensions import Self
import pytest

if TYPE_CHECKING:
    from pathlib import Path

    from _pytest.logging import LogCaptureFixture
    from pytest_mock import MockerFixture


def _get_fake_temp_dir(path: str) -> Any:
    """Return a fake temporary directory path."""
    class _FakeTempDir:
        def __init__(self, *args: Any, **kwargs: Any) -> None:
            self.path = path

        def __enter__(self) -> str:
            return self.path

        def __exit__(self, exc_type: object, exc_val: object, exc_tb: object) -> None:
            pass

        def open(self, mode: str, encoding: str | None = None) -> IO[Any]:
            return open(self.path, mode, encoding=encoding)  # noqa: PTH123

    return _FakeTempDir


def test_dotnet_restore_yields_expected_packages(mocker: MockerFixture, tmp_path: Path) -> None:
    mocker.patch('livecheck.special.dotnet.tempfile.TemporaryDirectory',
                 _get_fake_temp_dir(str(tmp_path)))
    mock_run = mocker.patch('livecheck.special.dotnet.sp.run')
    find_output = (f'{tmp_path}/Newtonsoft.Json@13.0.1\n'
                   f'{tmp_path}/SomeOther.Package@2.1.0\n'
                   f'{tmp_path}/microsoft.aspnetcore.app.host@9.0.0\n'
                   f'{tmp_path}/runtime.win@8.0.0\n'
                   f'{tmp_path}/NotAPackage\n')

    def run_side_effect(args: list[str], **kwargs: Any) -> Any:
        if args[0] == 'dotnet':
            return mocker.Mock()
        if args[0] == 'find':
            return mocker.Mock(stdout=find_output)
        raise ValueError

    mock_run.side_effect = run_side_effect

    project_file = tmp_path / 'proj.csproj'
    project_file.write_text('<Project></Project>')
    results = list(dotnet_restore(project_file))
    assert results == ['Newtonsoft.Json@13.0.1', 'SomeOther.Package@2.1.0']


def test_dotnet_restore_filters_correctly(mocker: MockerFixture, tmp_path: Path) -> None:
    mocker.patch('livecheck.special.dotnet.tempfile.TemporaryDirectory',
                 _get_fake_temp_dir(str(tmp_path)))
    mock_run = mocker.patch('livecheck.special.dotnet.sp.run')
    find_output = (f'{tmp_path}/microsoft.netcore.app.ref@9.0.0\n'
                   f'{tmp_path}/runtime.win@8.0.0\n'
                   f'{tmp_path}/Valid.Package@1.2.3\n')

    def run_side_effect(args: Any, **kwargs: Any) -> Any:
        if args[0] == 'dotnet':
            return mocker.Mock()
        if args[0] == 'find':
            return mocker.Mock(stdout=find_output)
        return None

    mock_run.side_effect = run_side_effect
    project_file = tmp_path / 'proj.csproj'
    project_file.write_text('<Project></Project>')
    results = list(dotnet_restore(project_file))
    assert results == ['Valid.Package@1.2.3']


def test_dotnet_restore_handles_no_packages(mocker: MockerFixture, tmp_path: Path) -> None:
    mock_run = mocker.patch('subprocess.run')
    find_output = ''

    def run_side_effect(args: Any, **kwargs: Any) -> Any:
        if args[0] == 'dotnet':
            return mocker.Mock()
        if args[0] == 'find':
            return mocker.Mock(stdout=find_output)
        return None

    mock_run.side_effect = run_side_effect
    project_file = tmp_path / 'proj.csproj'
    project_file.write_text('<Project></Project>')
    results = list(dotnet_restore(project_file))
    assert results == []


def test_dotnet_restore_raises_on_dotnet_failure(mocker: MockerFixture, tmp_path: Path) -> None:
    # Simulate dotnet restore failing
    def run_side_effect(args: Any, **kwargs: Any) -> Any:
        if args[0] == 'dotnet':
            msg = 'dotnet restore failed'
            raise RuntimeError(msg)
        if args[0] == 'find':
            return mocker.Mock(stdout='')
        return None

    mocker.patch('subprocess.run', side_effect=run_side_effect)
    project_file = tmp_path / 'proj.csproj'
    project_file.write_text('<Project></Project>')
    with pytest.raises(RuntimeError):
        list(dotnet_restore(project_file))


class _DummyTempFile:
    def __init__(self, path: str) -> None:
        self.path = path

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type: object, exc_val: object, exc_tb: object) -> None:
        pass

    def open(self, mode: str, encoding: str | None = None) -> IO[Any]:
        return open(self.path, mode, encoding=encoding)  # noqa: PTH123


def test_update_dotnet_ebuild_updates_nugets_correctly(mocker: MockerFixture,
                                                       tmp_path: Path) -> None:
    ebuild_path = tmp_path / 'test.ebuild'
    project_name = 'proj.csproj'
    project_path = tmp_path / project_name
    project_path.write_text('<Project></Project>')
    ebuild_content = [
        'EAPI=8\n', 'NUGETS="\n', '    Old.Package@1.0.0"\n', 'SOME_OTHER_VAR="foo"\n'
    ]
    ebuild_path.write_text(''.join(ebuild_content))
    mocker.patch('livecheck.special.dotnet.search_ebuild', return_value=(str(tmp_path), None))
    mocker.patch('livecheck.special.dotnet.dotnet_restore',
                 return_value=iter(['New.Package@2.0.0', 'Another.Package@3.1.4']))
    update_dotnet_ebuild(str(ebuild_path), project_name)
    # Check that the ebuild file was updated with new NUGETS
    lines = ebuild_path.read_text().splitlines()
    assert any('NUGETS="' in line for line in lines)
    assert any('New.Package@2.0.0' in line for line in lines)
    assert any('Another.Package@3.1.4' in line for line in lines)
    assert not any('Old.Package@1.0.0' in line for line in lines)


def test_update_dotnet_ebuild_raises_no_nugets_found(mocker: MockerFixture, tmp_path: Path) -> None:
    ebuild_path = tmp_path / 'test.ebuild'
    ebuild_path.write_text('EAPI=8\nSOME_OTHER_VAR="foo"\n')
    mocker.patch('livecheck.special.dotnet.search_ebuild', return_value=(str(tmp_path), None))
    mocker.patch('livecheck.special.dotnet.dotnet_restore', return_value=iter([]))
    mocker.patch('livecheck.special.dotnet.EbuildTempFile', _DummyTempFile)
    proj_csproj = tmp_path / 'proj.csproj'
    proj_csproj.write_text('<Project></Project>')
    with pytest.raises(NoNugetsFound):
        update_dotnet_ebuild(str(ebuild_path), proj_csproj)


def test_update_dotnet_ebuild_raises_runtime_error(mocker: MockerFixture, tmp_path: Path) -> None:
    ebuild_path = tmp_path / 'test.ebuild'
    ebuild_path.write_text('EAPI=8\nNUGETS="\nfoo\nbar\nNUGETS="\n')
    mocker.patch('livecheck.special.dotnet.search_ebuild', return_value=(str(tmp_path), None))
    mocker.patch('livecheck.special.dotnet.dotnet_restore', return_value=iter([]))
    proj_csproj = tmp_path / 'proj.csproj'
    proj_csproj.write_text('<Project></Project>')
    with pytest.raises(RuntimeError):
        update_dotnet_ebuild(str(ebuild_path), proj_csproj)


def test_update_dotnet_ebuild_raises_no_nugets_ending(mocker: MockerFixture,
                                                      tmp_path: Path) -> None:
    ebuild_path = tmp_path / 'test.ebuild'
    ebuild_path.write_text('EAPI=8\nNUGETS="\nfoo\nbar\n')
    mocker.patch('livecheck.special.dotnet.search_ebuild', return_value=(str(tmp_path), None))
    mocker.patch('livecheck.special.dotnet.dotnet_restore', return_value=iter([]))
    proj_csproj = tmp_path / 'proj.csproj'
    proj_csproj.write_text('<Project></Project>')
    with pytest.raises(NoNugetsEnding):
        update_dotnet_ebuild(str(ebuild_path), proj_csproj)


def test_update_dotnet_ebuild_returns_if_no_dotnet_path(mocker: MockerFixture,
                                                        tmp_path: Path) -> None:
    ebuild_path = tmp_path / 'test.ebuild'
    ebuild_path.write_text('EAPI=8\nNUGETS="\n    Old.Package@1.0.0"\n')
    mocker.patch('livecheck.special.dotnet.search_ebuild', return_value=(None, None))
    update_dotnet_ebuild(str(ebuild_path), 'proj.csproj')
    assert 'Old.Package@1.0.0' in ebuild_path.read_text()


def test_check_dotnet_requirements_returns_true_for_valid_version(mocker: MockerFixture) -> None:
    # Mock check_program to return True (dotnet >= 10.0.0)
    mocker.patch('livecheck.special.dotnet.check_program', return_value=True)
    assert check_dotnet_requirements() is True


def test_check_dotnet_requirements_returns_false_for_invalid_version(mocker: MockerFixture) -> None:
    # Mock check_program to return False (dotnet < 10.0.0 or not installed)
    mocker.patch('livecheck.special.dotnet.check_program', return_value=False)
    assert check_dotnet_requirements() is False


def test_check_dotnet_requirements_logs_error_on_failure(mocker: MockerFixture,
                                                         caplog: LogCaptureFixture) -> None:
    # Mock check_program to return False
    mocker.patch('livecheck.special.dotnet.check_program', return_value=False)
    with caplog.at_level('ERROR'):
        result = check_dotnet_requirements()
    assert result is False
    assert any(
        'dotnet is not installed or version is less than 9.0.0.' in m for m in caplog.messages)



from __future__ import annotations

from typing import TYPE_CHECKING

from livecheck.special.pypi import (
    extract_project,
    get_latest_pypi_metadata,
    get_latest_pypi_package,
    get_url,
    is_pypi,
)
import pytest

if TYPE_CHECKING:
    from collections.abc import Collection, Mapping

    from pytest_mock import MockerFixture

test_cases = [
    {
        'url': 'https://pypi.org/project/source/s/someproject/1.0.0/',
        'expected': 'someproject'
    },
    {
        'url': 'https://pypi.org/project/someproject-1.0.0/',
        'expected': ''
    },
    {
        'url':
            'https://files.pythonhosted.org/packages/source/s/someproject/someproject-1.0.0.tar.gz',
        'expected':
            'someproject'
    },
    {
        'url': 'https://pypi.io/packages/source/s/someproject/someproject-1.0.0.tar.gz',
        'expected': 'someproject'
    },
    {
        'url': 'https://example.com/project/someproject-1.0.0/',
        'expected': ''
    },
    {
        'url': 'https://pypi.org/project/someproject/',
        'expected': ''
    },
    {
        'url': 'https://pypi/project/source/s/someproject/1.0.0/',
        'expected': 'someproject'
    },
    {
        'url': 'https://pypi.io/project/source/s/someproject/1.0.0/',
        'expected': 'someproject'
    },
    {
        'url': 'https://1pypi.io/project/source/someproject/1.0.0/',
        'expected': ''
    },
    {
        'url': 'https://www.pythonhosted.org/project/source/someproject/1.0.0/',
        'expected': ''
    },
    {
        'url': 'mirror://pypi/project/source/s/someproject/1.0.0/',
        'expected': 'someproject'
    },
    {
        'url':
            'https://files.pythonhosted.org/packages/15/1f/ca74b65b19798895d63a6e92874162f44233467c9e7c1ed8afd19016ebe9/chevron-0.14.0.tar.gz',
        'expected':
            'chevron'
    },
]


@pytest.mark.parametrize('case', test_cases)
def test_extract_project(case: dict[str, str]) -> None:
    assert extract_project(case['url']) == case['expected']


@pytest.mark.parametrize(
    ('ext', 'item', 'expected'),
    [
        # Test: url with matching extension
        (
            '.tar.gz',
            [
                {
                    'url':
                        'https://files.pythonhosted.org/packages/source/s/someproject/someproject-1.0.0.tar.gz'
                },
                {
                    'url':
                        'https://files.pythonhosted.org/packages/source/s/someproject/someproject-1.0.0.zip'
                },
            ],
            'https://files.pythonhosted.org/packages/source/s/someproject/someproject-1.0.0.tar.gz',
        ),
        # Test: no url with matching extension, but one with archive extension
        (
            '.whl',
            [
                {
                    'url':
                        'https://files.pythonhosted.org/packages/source/s/someproject/someproject-1.0.0.tar.gz'
                },
                {
                    'url':
                        'https://files.pythonhosted.org/packages/source/s/someproject/someproject-1.0.0.zip'
                },
            ],
            'https://files.pythonhosted.org/packages/source/s/someproject/someproject-1.0.0.tar.gz',
        ),
        # Test: no url with matching extension, no archive extension
        (
            '.exe',
            [
                {
                    'url':
                        'https://files.pythonhosted.org/packages/source/s/someproject/someproject-1.0.0.txt'
                },
            ],
            '',
        ),
        # Test: empty item list
        (
            '.tar.gz',
            [],
            '',
        ),
    ])
def test_get_url(mocker: MockerFixture, ext: str, item: Collection[Mapping[str, str]],
                 expected: str) -> None:
    def fake_get_archive_extension(url: str) -> str:
        if url.endswith(('.tar.gz', '.zip')):
            return '.tar.gz' if url.endswith('.tar.gz') else '.zip'
        return ''

    mocker.patch('livecheck.special.pypi.get_archive_extension',
                 side_effect=fake_get_archive_extension)
    assert get_url(ext, item) == expected


@pytest.mark.parametrize(('src_uri', 'project_name', 'latest_version', 'latest_url'), [
    ('https://files.pythonhosted.org/packages/source/s/someproject/someproject-1.0.0.tar.gz',
     'someproject', '2.0.0',
     'https://files.pythonhosted.org/packages/source/s/someproject/someproject-2.0.0.tar.gz'),
    ('https://files.pythonhosted.org/packages/source/s/another/another-0.1.0.zip', 'another',
     '0.2.0', 'https://files.pythonhosted.org/packages/source/s/another/another-0.2.0.zip'),
])
def test_get_latest_pypi_package_success(mocker: MockerFixture, src_uri: str, project_name: str,
                                         latest_version: str, latest_url: str) -> None:
    mocker.patch('livecheck.special.pypi.extract_project', return_value=project_name)
    ext = '.tar.gz' if src_uri.endswith('.tar.gz') else '.zip'
    mocker.patch('livecheck.special.pypi.get_archive_extension', return_value=ext)
    releases = {latest_version: [{'url': latest_url}], '1.0.0': [{'url': src_uri}]}
    mock_response = mocker.Mock()
    mock_response.json.return_value = {'releases': releases}
    mocker.patch('livecheck.special.pypi.get_content', return_value=mock_response)
    mocker.patch('livecheck.special.pypi.get_last_version',
                 return_value={
                     'version': latest_version,
                     'url': latest_url
                 })

    version, url = get_latest_pypi_package(src_uri, 'dummy.ebuild', mocker.Mock())
    assert version == latest_version
    assert url == latest_url


def test_get_latest_pypi_package_no_content(mocker: MockerFixture) -> None:
    mocker.patch('livecheck.special.pypi.extract_project', return_value='someproject')
    mocker.patch('livecheck.special.pypi.get_archive_extension', return_value='.tar.gz')
    mocker.patch('livecheck.special.pypi.get_content', return_value=None)
    version, url = get_latest_pypi_package(
        'https://files.pythonhosted.org/packages/source/s/someproject/someproject-1.0.0.tar.gz',
        'dummy.ebuild', mocker.Mock())
    assert not version
    assert not url


def test_get_latest_pypi_package_no_last_version(mocker: MockerFixture) -> None:
    mocker.patch('livecheck.special.pypi.extract_project', return_value='someproject')
    mocker.patch('livecheck.special.pypi.get_archive_extension', return_value='.tar.gz')
    mock_response = mocker.Mock()
    mock_response.json.return_value = {'releases': {'1.0.0': [{'url': 'url'}]}}
    mocker.patch('livecheck.special.pypi.get_content', return_value=mock_response)
    mocker.patch('livecheck.special.pypi.get_last_version', return_value=None)
    version, url = get_latest_pypi_package(
        'https://files.pythonhosted.org/packages/source/s/someproject/someproject-1.0.0.tar.gz',
        'dummy.ebuild', mocker.Mock())
    assert not version
    assert not url


@pytest.mark.parametrize(('url', 'extract_project_return', 'expected'), [
    ('https://pypi.org/project/source/s/someproject/1.0.0/', 'someproject', True),
    ('https://pypi.org/project/someproject-1.0.0/', '', False),
    ('https://files.pythonhosted.org/packages/source/s/someproject/someproject-1.0.0.tar.gz',
     'someproject', True),
    ('https://example.com/project/someproject-1.0.0/', '', False),
    ('mirror://pypi/project/source/s/someproject/1.0.0/', 'someproject', True),
    ('https://pypi.io/project/source/s/someproject/1.0.0/', 'someproject', True),
    ('https://1pypi.io/project/source/someproject/1.0.0/', '', False),
    ('', '', False),
])
def test_is_pypi(mocker: MockerFixture, url: str, extract_project_return: str,
                 expected: bool) -> None:  # noqa: FBT001
    mocker.patch('livecheck.special.pypi.extract_project', return_value=extract_project_return)
    assert is_pypi(url) is expected


@pytest.mark.parametrize(('remote', 'ebuild', 'latest_version', 'latest_url'), [
    ('someproject', 'dummy.ebuild', '2.0.0',
     'https://files.pythonhosted.org/packages/source/s/someproject/someproject-2.0.0.tar.gz'),
    ('another', 'another.ebuild', '0.2.0',
     'https://files.pythonhosted.org/packages/source/s/another/another-0.2.0.zip'),
])
def test_get_latest_pypi_metadata_success(mocker: MockerFixture, remote: str, ebuild: str,
                                          latest_version: str, latest_url: str) -> None:
    mocker.patch('livecheck.special.pypi.get_latest_pypi_package2',
                 return_value=(latest_version, latest_url))
    version, url = get_latest_pypi_metadata(remote, ebuild, mocker.Mock())
    assert version == latest_version
    assert url == latest_url


def test_get_latest_pypi_metadata_empty(mocker: MockerFixture) -> None:
    mocker.patch('livecheck.special.pypi.get_latest_pypi_package2', return_value=('', ''))
    version, url = get_latest_pypi_metadata('remote', 'ebuild', mocker.Mock())
    assert not version
    assert not url



from __future__ import annotations

from typing import TYPE_CHECKING

from livecheck.special.davinci import get_latest_davinci_package
import pytest

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_get_latest_davinci_package_success(mocker: MockerFixture) -> None:
    mock_response = mocker.Mock()
    mock_response.json.return_value = {'linux': {'major': 18, 'minor': 5, 'releaseNum': 1}}
    mock_get_content = mocker.patch('livecheck.special.davinci.get_content',
                                    return_value=mock_response)
    result = get_latest_davinci_package('davinci')
    assert result == '18.5.1'
    mock_get_content.assert_called_once_with(
        'https://www.blackmagicdesign.com/api/support/latest-stable-version/davinci/linux')


def test_get_latest_davinci_package_no_content(mocker: MockerFixture) -> None:
    mocker.patch('livecheck.special.davinci.get_content', return_value=None)
    result = get_latest_davinci_package('davinci')
    assert not result


def test_get_latest_davinci_package_missing_linux_key(mocker: MockerFixture) -> None:
    mock_response = mocker.Mock()
    mock_response.json.return_value = {}
    mocker.patch('livecheck.special.davinci.get_content', return_value=mock_response)
    with pytest.raises(KeyError):
        get_latest_davinci_package('davinci')


def test_get_latest_davinci_package_partial_linux_data(mocker: MockerFixture) -> None:
    mock_response = mocker.Mock()
    mock_response.json.return_value = {'linux': {'major': 18, 'minor': 5}}
    mocker.patch('livecheck.special.davinci.get_content', return_value=mock_response)
    with pytest.raises(KeyError):
        get_latest_davinci_package('davinci')



from __future__ import annotations

from typing import TYPE_CHECKING

from livecheck.special.jetbrains import (
    get_latest_jetbrains_package,
    is_jetbrains,
    update_jetbrains_ebuild,
)

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_mock import MockerFixture


def test_get_latest_jetbrains_package_returns_latest(mocker: MockerFixture) -> None:
    mock_settings = mocker.Mock()
    mock_settings.is_devel.return_value = False

    # Mock catpkg_catpkgsplit to return expected values
    mocker.patch('livecheck.special.jetbrains.catpkg_catpkgsplit',
                 return_value=('dev-util', None, 'pycharm-community', None))

    # Mock get_content to return a mock response with .json()
    mock_response = mocker.Mock()
    mock_response.json.return_value = [{
        'name':
            'PyCharm Community Edition',
        'releases': [
            {
                'type': 'release',
                'version': '2023.1',
                'downloads': {
                    'linux': {}
                }
            },
            {
                'type': 'eap',
                'version': '2023.2',
                'downloads': {
                    'linux': {}
                }
            },
        ]
    }]
    mocker.patch('livecheck.special.jetbrains.get_content', return_value=mock_response)

    # Mock get_last_version to return the latest version dict
    mocker.patch('livecheck.special.jetbrains.get_last_version', return_value={'tag': '2023.1'})

    result = get_latest_jetbrains_package('dev-util/pycharm-community', mock_settings)
    assert result == '2023.1'


def test_get_latest_jetbrains_package_no_content(mocker: MockerFixture) -> None:
    mock_settings = mocker.Mock()
    mocker.patch('livecheck.special.jetbrains.catpkg_catpkgsplit',
                 return_value=('dev-util', None, 'clion', None))
    mocker.patch('livecheck.special.jetbrains.get_content', return_value=None)
    result = get_latest_jetbrains_package('dev-util/clion', mock_settings)
    assert not result


def test_get_latest_jetbrains_package_no_matching_product(mocker: MockerFixture) -> None:
    mock_settings = mocker.Mock()
    mock_settings.is_devel.return_value = False
    mocker.patch('livecheck.special.jetbrains.catpkg_catpkgsplit',
                 return_value=('dev-util', None, 'unknown-product', None))
    mock_response = mocker.Mock()
    mock_response.json.return_value = [{'name': 'PyCharm Community Edition', 'releases': []}]
    mocker.patch('livecheck.special.jetbrains.get_content', return_value=mock_response)
    mocker.patch('livecheck.special.jetbrains.get_last_version', return_value=None)
    result = get_latest_jetbrains_package('dev-util/unknown-product', mock_settings)
    assert not result


def test_get_latest_jetbrains_package_skips_eap_and_rc_if_not_devel(mocker: MockerFixture) -> None:
    mock_settings = mocker.Mock()
    mock_settings.is_devel.return_value = False
    mocker.patch('livecheck.special.jetbrains.catpkg_catpkgsplit',
                 return_value=('dev-util', None, 'clion', None))
    mock_response = mocker.Mock()
    mock_response.json.return_value = [{
        'name':
            'CLion',
        'releases': [
            {
                'type': 'eap',
                'version': '2023.2',
                'downloads': {
                    'linux': {}
                }
            },
            {
                'type': 'rc',
                'version': '2023.3',
                'downloads': {
                    'linux': {}
                }
            },
            {
                'type': 'release',
                'version': '2023.1',
                'downloads': {
                    'linux': {}
                }
            },
        ]
    }]
    mocker.patch('livecheck.special.jetbrains.get_content', return_value=mock_response)
    mocker.patch('livecheck.special.jetbrains.get_last_version', return_value={'tag': '2023.1'})
    result = get_latest_jetbrains_package('dev-util/clion', mock_settings)
    assert result == '2023.1'


def test_get_latest_jetbrains_package_includes_eap_and_rc_if_devel(mocker: MockerFixture) -> None:
    mock_settings = mocker.Mock()
    mock_settings.is_devel.return_value = True
    mocker.patch('livecheck.special.jetbrains.catpkg_catpkgsplit',
                 return_value=('dev-util', None, 'clion', None))
    mock_response = mocker.Mock()
    mock_response.json.return_value = [{
        'name':
            'CLion',
        'releases': [
            {
                'type': 'eap',
                'version': '2023.2',
                'downloads': {
                    'linux': {}
                }
            },
            {
                'type': 'rc',
                'version': '2023.3',
                'downloads': {
                    'linux': {}
                }
            },
            {
                'type': 'release',
                'version': '2023.1',
                'downloads': {
                    'linux': {}
                }
            },
            {
                'type': 'release',
                'version': '2023.1',
                'downloads': {
                    'mac': {}
                }
            },
        ]
    }]
    mocker.patch('livecheck.special.jetbrains.get_content', return_value=mock_response)
    mocker.patch('livecheck.special.jetbrains.get_last_version', return_value={'tag': '2023.3'})
    result = get_latest_jetbrains_package('dev-util/clion', mock_settings)
    assert result == '2023.3'


def test_update_jetbrains_ebuild_updates_my_pv_line(mocker: MockerFixture, tmp_path: Path) -> None:
    # Setup
    ebuild_path = tmp_path / 'fake-ebuild.ebuild'
    fake_version = '2024.1'
    fake_package_path = f'/some/path/product-{fake_version}'
    # Patch search_ebuild to return a path containing the version
    mocker.patch('livecheck.special.jetbrains.search_ebuild',
                 return_value=(fake_package_path, None))
    # Patch EbuildTempFile context manager
    mock_temp_file = mocker.MagicMock()
    mock_temp_file.__enter__.return_value = mock_temp_file
    mock_temp_file.__exit__.return_value = None
    mocker.patch('livecheck.special.jetbrains.EbuildTempFile', return_value=mock_temp_file)
    # Patch open for both temp_file and Path(ebuild)
    mock_write = mocker.mock_open()
    mock_read = mocker.mock_open(read_data='MY_PV="old"\nSOME=other\n')
    mocker.patch('pathlib.Path.open', mock_read)
    mock_temp_file.open = mock_write
    # Run
    update_jetbrains_ebuild(str(ebuild_path))
    # Assert
    handle = mock_write()
    handle.write.assert_any_call(f'MY_PV="{fake_version}"\n')
    handle.write.assert_any_call('SOME=other\n')


def test_update_jetbrains_ebuild_no_version_found(mocker: MockerFixture, tmp_path: Path) -> None:
    ebuild_path = tmp_path / 'fake-ebuild.ebuild'
    mocker.patch('livecheck.special.jetbrains.search_ebuild', return_value=('', None))
    mock_logger = mocker.patch('livecheck.special.jetbrains.logger')
    mocker.patch('livecheck.special.jetbrains.EbuildTempFile')
    update_jetbrains_ebuild(str(ebuild_path))
    mock_logger.warning.assert_called_once_with('No version found in the tar.gz file.')


def test_update_jetbrains_ebuild_handles_no_my_pv_line(mocker: MockerFixture,
                                                       tmp_path: Path) -> None:
    ebuild_path = tmp_path / 'fake-ebuild.ebuild'
    fake_version = '2024.2'
    fake_package_path = f'/some/path/product-{fake_version}'
    mocker.patch('livecheck.special.jetbrains.search_ebuild',
                 return_value=(fake_package_path, None))
    mock_temp_file = mocker.MagicMock()
    mock_temp_file.__enter__.return_value = mock_temp_file
    mock_temp_file.__exit__.return_value = None
    mocker.patch('livecheck.special.jetbrains.EbuildTempFile', return_value=mock_temp_file)
    mock_write = mocker.mock_open()
    mock_read = mocker.mock_open(read_data='SOME=other\n')
    mocker.patch('pathlib.Path.open', mock_read)
    mock_temp_file.open = mock_write
    update_jetbrains_ebuild(str(ebuild_path))
    handle = mock_write()
    handle.write.assert_any_call('SOME=other\n')


def test_is_jetbrains_true_for_download_url(mocker: MockerFixture) -> None:
    url = 'https://download.jetbrains.com/python/pycharm-community-2023.1.tar.gz'
    assert is_jetbrains(url)



"""JetBrains functions."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import urlparse
import logging

from livecheck.utils import get_content
from livecheck.utils.portage import catpkg_catpkgsplit, get_last_version

from .utils import EbuildTempFile, search_ebuild

if TYPE_CHECKING:
    from livecheck.settings import LivecheckSettings

__all__ = ('get_latest_jetbrains_package', 'is_jetbrains', 'update_jetbrains_ebuild')

JETBRAINS_TAG_URL = 'https://data.services.jetbrains.com/products'
logger = logging.getLogger(__name__)


def get_latest_jetbrains_package(ebuild: str, settings: LivecheckSettings) -> str:
    """Get the latest version of a JetBrains package."""
    product_name = {
        'phpstorm': 'PhpStorm',
        'pycharm-community': 'PyCharm Community Edition',
        'pycharm-professional': 'PyCharm Professional Edition',
        'idea-community': 'IntelliJ IDEA Community Edition',
        'clion': 'CLion',
        'goland': 'GoLand',
    }

    catpkg, _, product_code, _ = catpkg_catpkgsplit(ebuild)

    if not (r := get_content(JETBRAINS_TAG_URL)):
        return ''

    product_code = product_name.get(product_code, product_code)

    results: list[dict[str, str]] = []
    for product in r.json():
        if product['name'] == product_code:
            for release in product['releases']:
                if (release['type'] == 'eap'
                        or release['type'] == 'rc') and not settings.is_devel(catpkg):
                    continue
                if 'linux' in release.get('downloads', ''):
                    results.append({'tag': release['version']})

    if last_version := get_last_version(results, '', ebuild, settings):
        return last_version['tag']

    return ''


def update_jetbrains_ebuild(ebuild: str) -> None:
    """Update a JetBrains package ebuild."""
    package_path, _ = search_ebuild(str(ebuild), 'product-info.json')
    if not (version := package_path.split('/')[-1]):
        logger.warning('No version found in the tar.gz file.')
        return

    version = version.split('-', 1)[-1]

    with EbuildTempFile(ebuild) as temp_file, temp_file.open(
            'w', encoding='utf-8') as tf, Path(ebuild).open('r', encoding='utf-8') as f:
        for line in f:
            if line.startswith('MY_PV='):
                logger.debug('Found MY_PV= line.')
                tf.write(f'MY_PV="{version}"\n')
            else:
                tf.write(line)


def is_jetbrains(url: str) -> bool:
    """Check if the URL is a JetBrains download URL."""
    return urlparse(url).netloc == 'download.jetbrains.com'



"""Type assertions."""
from __future__ import annotations

from typing import TypeVar

T = TypeVar('T')


def assert_not_none(x: T | None) -> T:  # pragma: no cover
    """Assert that the value is not None."""
    assert x is not None
    return x



# ruff: noqa: S108
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from livecheck.special.yarn import (
    check_yarn_requirements,
    create_project,
    update_yarn_ebuild,
    yarn_pkgs,
)
import pytest

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_create_project_runs_yarn_commands(mocker: MockerFixture) -> None:
    mock_run = mocker.patch('livecheck.special.yarn.sp.run')
    mock_get_project_path = mocker.patch('livecheck.special.yarn.get_project_path',
                                         return_value=Path('/tmp/project'))
    base_package = 'foo'
    yarn_packages = {'bar', 'baz'}
    result = create_project(base_package, yarn_packages)
    assert result == Path('/tmp/project')
    assert mock_get_project_path.call_count == 1
    assert mock_run.call_count == 3
    assert mock_run.call_args_list[0][0][0][:4] == ('yarn', 'config', 'set', 'ignore-engines')
    add_args = mock_run.call_args_list[1][0][0]
    assert add_args[0:2] == ('yarn', 'add')
    assert base_package in add_args
    assert all(pkg in add_args for pkg in yarn_packages)
    assert mock_run.call_args_list[2][0][0][:2] == ('yarn', 'upgrade')


def test_yarn_pkgs_returns_expected_packages(mocker: MockerFixture) -> None:
    mock_project_path = mocker.MagicMock()
    mocker.patch('livecheck.special.yarn.sp.run')
    mocker.patch('livecheck.special.yarn.json.loads',
                 return_value={
                     'foo@^1.0.0': {
                         'version': '1.2.3',
                         'resolved': 'https://example.com/foo-1.2.3.tgz',
                         'integrity': 'sha512-...',
                         'dependencies': {
                             'bar': '^2.0.0'
                         }
                     },
                     '@scope/bar@^2.0.0': {
                         'version': '2.3.4',
                         'resolved': 'https://example.com/bar-2.3.4.tgz',
                         'integrity': 'sha512-...',
                     },
                     'baz-cjs@^3.0.0': {
                         'version': '3.4.5',
                         'resolved': 'https://example.com/baz-3.4.5.tgz',
                         'integrity': 'sha512-...',
                     }
                 })
    pkgs = list(yarn_pkgs(mock_project_path))
    assert 'foo-1.2.3' in pkgs
    assert '@scope/bar-2.3.4' in pkgs
    assert all(not pkg.startswith('baz-cjs') for pkg in pkgs)
    assert len(pkgs) == 2


def test_update_yarn_ebuild_replaces_yarn_pkgs_section(mocker: MockerFixture) -> None:
    ebuild_path = '/tmp/test.ebuild'
    yarn_base_package = 'foo'
    pkg = 'cat/foo'
    yarn_packages = {'bar', 'baz'}
    mocker.patch('pathlib.Path.chmod')
    mock_create_project = mocker.patch('livecheck.special.yarn.create_project',
                                       return_value=Path('/tmp/project'))
    mock_yarn_pkgs = mocker.patch('livecheck.special.yarn.yarn_pkgs',
                                  return_value=['foo-1.2.3', 'bar-2.3.4'])
    mock_copyfile = mocker.patch('livecheck.special.yarn.copyfile')
    mock_ebuild_temp_file = mocker.patch('livecheck.special.yarn.EbuildTempFile')
    mock_temp_file = mock_ebuild_temp_file.return_value
    mock_temp_file.__enter__.return_value = mock_temp_file
    mock_temp_file.open.return_value.__enter__.return_value = mocker.MagicMock()
    mock_temp_file.exists.return_value = True
    mock_temp_file.stat.return_value.st_size = 10
    ebuild_content = ['EAPI=8\n', 'YARN_PKGS=(\n', '    old-pkg-0.1.0\n', ')\n', 'SRC_URI=...\n']
    mocker.patch('pathlib.Path.open', mocker.mock_open(read_data=''.join(ebuild_content)))
    tf_mock = mock_temp_file.open.return_value.__enter__.return_value
    update_yarn_ebuild(ebuild_path, yarn_base_package, pkg, yarn_packages)
    mock_create_project.assert_called_once_with(yarn_base_package, yarn_packages)
    mock_yarn_pkgs.assert_called_once()
    mock_ebuild_temp_file.assert_called_once_with(ebuild_path)
    tf_mock.write.assert_any_call('YARN_PKGS=(\n')
    tf_mock.write.assert_any_call('\tbar-2.3.4\n')
    tf_mock.write.assert_any_call('\tfoo-1.2.3\n')
    tf_mock.write.assert_any_call(')\n')
    mock_copyfile.assert_any_call(
        Path('/tmp/project') / 'package.json',
        Path('/tmp/test.ebuild').parent / 'files' / 'foo-package.json')
    mock_copyfile.assert_any_call(
        Path('/tmp/project') / 'yarn.lock',
        Path('/tmp/test.ebuild').parent / 'files' / 'foo-yarn.lock')


def test_update_yarn_ebuild_raises_on_malformed_section(mocker: MockerFixture) -> None:
    ebuild_path = '/tmp/test.ebuild'
    yarn_base_package = 'foo'
    pkg = 'cat/foo'
    yarn_packages = None
    mocker.patch('livecheck.special.yarn.create_project', return_value=Path('/tmp/project'))
    mocker.patch('livecheck.special.yarn.yarn_pkgs', return_value=['foo-1.2.3'])
    mocker.patch('livecheck.special.yarn.copyfile')
    mock_ebuild_temp_file = mocker.patch('livecheck.special.yarn.EbuildTempFile')
    mock_temp_file = mock_ebuild_temp_file.return_value
    mock_temp_file.__enter__.return_value = mock_temp_file
    mock_temp_file.open.return_value.__enter__.return_value = mocker.MagicMock()
    mock_temp_file.exists.return_value = False
    mock_temp_file.stat.return_value.st_size = 0
    ebuild_content = ['YARN_PKGS=(\n', 'YARN_PKGS=(\n', ')\n']
    mocker.patch('pathlib.Path.open', mocker.mock_open(read_data=''.join(ebuild_content)))
    with pytest.raises(RuntimeError):
        update_yarn_ebuild(ebuild_path, yarn_base_package, pkg, yarn_packages)


def test_check_yarn_requirements_all_installed(mocker: MockerFixture) -> None:
    mock_check_program = mocker.patch('livecheck.special.yarn.check_program',
                                      side_effect=[True, True])
    assert check_yarn_requirements() is True
    assert mock_check_program.call_count == 2
    mock_check_program.assert_any_call('yarn', ['--version'])
    mock_check_program.assert_any_call('node', ['--version'])


def test_check_yarn_requirements_yarn_missing(mocker: MockerFixture) -> None:
    mock_check_program = mocker.patch('livecheck.special.yarn.check_program',
                                      side_effect=[False, True])
    mock_logger = mocker.patch('livecheck.special.yarn.logger')
    assert check_yarn_requirements() is False
    mock_logger.error.assert_called_once_with('yarn is not installed')
    assert mock_check_program.call_count == 1


def test_check_yarn_requirements_node_missing(mocker: MockerFixture) -> None:
    mock_check_program = mocker.patch('livecheck.special.yarn.check_program',
                                      side_effect=[True, False])
    mock_logger = mocker.patch('livecheck.special.yarn.logger')
    assert check_yarn_requirements() is False
    mock_logger.error.assert_called_once_with('yarn is not installed')
    assert mock_check_program.call_count == 2



"""Bitbucket functions."""
from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import urlparse

from livecheck.utils import get_content, is_sha
from livecheck.utils.portage import get_last_version

from .utils import get_archive_extension, log_unhandled_commit

if TYPE_CHECKING:
    from livecheck.settings import LivecheckSettings

__all__ = ('BITBUCKET_METADATA', 'get_latest_bitbucket', 'get_latest_bitbucket_metadata',
           'get_latest_bitbucket_package', 'is_bitbucket')

# doc: https://developer.atlassian.com/cloud/bitbucket/rest/api-group-refs/#api-repositories-workspace-repo-slug-refs-tags-get
BITBUCKET_TAG_URL = 'https://api.bitbucket.org/2.0/repositories/%s/%s/refs/tags'
# doc: https://developer.atlassian.com/cloud/bitbucket/rest/api-group-downloads/#api-repositories-workspace-repo-slug-downloads-get
BITBUCKET_DOWNLOAD_URL = 'https://api.bitbucket.org/2.0/repositories/%s/%s/downloads'
BITBUCKET_METADATA = 'bitbucket'
MAX_ITERATIONS = 4


def extract_workspace_and_repository(url: str) -> tuple[str, str]:
    parsed = urlparse(url)

    if (parsed.netloc != 'bitbucket.org'
            or len(parsed.path.strip('/').split('/')) < 2):  # noqa: PLR2004
        return '', ''

    workspace, repository = parsed.path.strip('/').split('/')[:2]

    return workspace, repository.replace('.git', '')


def get_latest_bitbucket_package(url: str, cpv: str,
                                 settings: LivecheckSettings) -> tuple[str, str]:
    """Get the latest version of a Bitbucket package."""
    workspace, repository = extract_workspace_and_repository(url)

    url = BITBUCKET_TAG_URL % (workspace, repository)

    if not (tags_response := get_content(url)):
        return '', ''

    results: list[dict[str, str]] = [{
        'tag': tag.get('name', ''),
        'id': tag.get('target', {}).get('hash', '')
    } for tag in tags_response.json().get('values', [])]

    # The tag may not be created and you need to know the downloads
    # for the latest versions
    url = BITBUCKET_DOWNLOAD_URL % (workspace, repository)
    iteration_count = 0

    while url and iteration_count < MAX_ITERATIONS:
        if not (r := get_content(url)).ok:
            break
        data = r.json()

        results.extend({
            'tag': item.get('name', ''),
            'id': ''
        } for item in data.get('values', []) if get_archive_extension(item.get('name')))

        url = data.get('next')
        iteration_count += 1

    if last_version := get_last_version(results, repository, cpv, settings):
        return last_version['version'], last_version['id']

    return '', ''


def get_latest_bitbucket(url: str, cpv: str, settings: LivecheckSettings, *,
                         force_sha: bool) -> tuple[str, str, str]:
    """Get the latest version of a Bitbucket package."""
    last_version = top_hash = hash_date = ''

    if is_sha(urlparse(url).path):
        log_unhandled_commit(cpv, url)
    else:
        last_version, top_hash = get_latest_bitbucket_package(url, cpv, settings)
        if not force_sha:
            top_hash = ''

    return last_version, top_hash, hash_date


def is_bitbucket(url: str) -> bool:
    """Check if the URL is a Bitbucket repository."""
    return bool(extract_workspace_and_repository(url)[0])


def get_latest_bitbucket_metadata(remote: str, cpv: str,
                                  settings: LivecheckSettings) -> tuple[str, str]:
    """Get the latest version of a Bitbucket package from metadata."""
    return get_latest_bitbucket_package(f'https://bitbucket.org/{remote}', cpv, settings)



from __future__ import annotations

from typing import TYPE_CHECKING

from livecheck.special.directory import get_latest_directory_package

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_get_latest_directory_package_returns_latest(mocker: MockerFixture) -> None:
    # Arrange
    url = 'https://example.com/packages/foo-1.0.tar.gz'
    ebuild = 'foo-1.0.ebuild'
    settings = mocker.Mock()
    html = """<html>
      <body>
        <a href="foo-1.0.tar.gz">foo-1.0.tar.gz</a>
        <a href="foo-2.0.tar.gz">foo-2.0.tar.gz</a>
        <a href="foo-1.5.tar.gz">foo-1.5.tar.gz</a>
        <a href="bar-1.0.tar.gz">bar-1.0.tar.gz</a>
        <a href="">bad</a>
      </body>
    </html>"""
    mock_response = mocker.Mock()
    mock_response.text = html
    mocker.patch('livecheck.special.directory.get_content', return_value=mock_response)
    mocker.patch('livecheck.special.directory.get_archive_extension',
                 side_effect=lambda x: x.endswith('.tar.gz'))
    last_version = {'version': '2.0', 'url': '/packages/foo-2.0.tar.gz'}
    mocker.patch('livecheck.special.directory.get_last_version', return_value=last_version)
    version, file_url = get_latest_directory_package(url, ebuild, settings)
    assert version == '2.0'
    assert file_url == '/packages/foo-2.0.tar.gz'


def test_get_latest_directory_get_last_version_falsy(mocker: MockerFixture) -> None:
    url = 'https://example.com/packages/foo-1.0.tar.gz'
    ebuild = 'foo-1.0.ebuild'
    settings = mocker.Mock()
    html = '<html><body></body></html>'
    mock_response = mocker.Mock()
    mock_response.text = html
    mocker.patch('livecheck.special.directory.get_content', return_value=mock_response)
    mocker.patch('livecheck.special.directory.get_archive_extension', return_value=True)
    mocker.patch('livecheck.special.directory.get_last_version', return_value=None)
    version, file_url = get_latest_directory_package(url, ebuild, settings)
    assert not version
    assert not file_url


def test_get_latest_directory_package_no_results(mocker: MockerFixture) -> None:
    url = 'https://example.com/packages/foo-1.0.tar.gz'
    ebuild = 'foo-1.0.ebuild'
    settings = mocker.Mock()
    html = '<html><body></body></html>'
    mock_response = mocker.Mock()
    mock_response.text = html
    mocker.patch('livecheck.special.directory.get_content', return_value=mock_response)
    mocker.patch('livecheck.special.directory.get_archive_extension', return_value=False)
    mocker.patch('livecheck.special.directory.get_last_version', return_value=None)
    version, file_url = get_latest_directory_package(url, ebuild, settings)
    assert not version
    assert not file_url


def test_get_latest_directory_package_no_content(mocker: MockerFixture) -> None:
    url = 'https://example.com/packages/foo-1.0.tar.gz'
    ebuild = 'foo-1.0.ebuild'
    settings = mocker.Mock()
    mocker.patch('livecheck.special.directory.get_content', return_value=None)
    version, file_url = get_latest_directory_package(url, ebuild, settings)
    assert not version
    assert not file_url


def test_get_latest_directory_package_no_match_in_url(mocker: MockerFixture) -> None:
    url = 'https://example.com/packages/'
    ebuild = 'foo-1.0.ebuild'
    settings = mocker.Mock()
    version, file_url = get_latest_directory_package(url, ebuild, settings)
    assert not version
    assert not file_url



"""SourceForge utility functions."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import urlparse
import re

from defusedxml import ElementTree as ET  # noqa: N817
from livecheck.utils import get_content
from livecheck.utils.portage import get_last_version

from .utils import get_archive_extension

if TYPE_CHECKING:
    from livecheck.settings import LivecheckSettings

__all__ = ('SOURCEFORGE_METADATA', 'get_latest_sourceforge_metadata',
           'get_latest_sourceforge_package', 'is_sourceforge')

SOURCEFORGE_DOWNLOAD_URL = 'https://sourceforge.net/projects/%s/rss'
SOURCEFORGE_METADATA = 'sourceforge'


def extract_repository(url: str) -> str:
    parsed = urlparse(url)
    n = parsed.netloc
    if n in {'downloads.sourceforge.net', 'download.sourceforge.net', 'sf.net'}:
        path = parsed.path.split('/')
        if 'projects' in path[1] or 'project' in path[1]:
            return path[2]
        return path[1]

    if (m := re.match(r'^([^\.]+)\.(sf|sourceforge)\.(net|io|jp)$', n)):
        return m.group(1)

    return ''


def get_latest_sourceforge_package(src_uri: str, ebuild: str, settings: LivecheckSettings) -> str:
    """Get the latest version of a SourceForge package."""
    repository = extract_repository(src_uri)
    return get_latest_sourceforge_package2(repository, ebuild, settings)


def get_latest_sourceforge_package2(repository: str, ebuild: str,
                                    settings: LivecheckSettings) -> str:
    url = SOURCEFORGE_DOWNLOAD_URL % (repository)

    if not (r := get_content(url)):
        return ''

    results: list[dict[str, str]] = []
    for item in ET.fromstring(r.text).findall('.//item'):
        title = item.find('title')
        version = Path(title.text).name if title is not None and title.text else ''
        if version and get_archive_extension(version):
            results.append({'tag': version})

    if last_version := get_last_version(results, repository, ebuild, settings):
        return last_version['version']

    return ''


def is_sourceforge(url: str) -> bool:
    """Check if the URL is a SourceForge repository."""
    return bool(extract_repository(url))


def get_latest_sourceforge_metadata(remote: str, ebuild: str, settings: LivecheckSettings) -> str:
    """Get the latest version of a SourceForge package using metadata."""
    return get_latest_sourceforge_package2(remote, ebuild, settings)



"""Very special handlers."""
from __future__ import annotations

from functools import lru_cache
import logging
import re

from livecheck.utils import assert_not_none

from .github import get_latest_github_commit2

logger = logging.getLogger(__name__)


@lru_cache
def handle_glabels(s: str) -> str:
    """Handle glabels-qt versioning."""
    _, hash_date = get_latest_github_commit2('jimevins', 'glabels-qt', 'master')
    if not hash_date:
        return s
    return '3.99_p' + hash_date


def handle_re(s: str) -> str:
    """Handle re3, reVC, and reLCS versioning."""
    return re.sub(r'^re(3|VC|LCS)_v?', '', s)


def handle_cython_post_suffix(s: str) -> str:
    """Handle Cython ``.post`` suffix."""
    return s.replace('.post', '.')


OUTFOX_MAXSPLIT = 2


def handle_outfox(s: str) -> str:
    """Handle Outfox versioning."""
    x = re.split(r'-pre(?:0+)?', s, maxsplit=OUTFOX_MAXSPLIT)
    if len(x) == OUTFOX_MAXSPLIT:
        return f'{x[0]}_p{x[1]}'
    return x[0]


def handle_outfox_serenity(s: str) -> str:
    """Handle Outfox Serenity versioning."""
    return s.replace('s', '.')


def handle_bsnes_hd(s: str) -> str:
    """Handle bsnes-hd versioning."""
    logger.debug('handle_bsnes_hd() <- "%s"', s)
    major, minor = assert_not_none(re.match(r'^beta_(\d+)_(\d+(?:h\d+)?)', s)).groups()
    minor = re.sub(r'h\d+', '', minor)
    ret = f'{major}.{minor}_beta'
    logger.debug('handle_bsnes_hd() -> "%s"', ret)
    return ret


def handle_pl(s: str) -> str:
    r"""Handle ``-pl\d+`` versioning."""
    logger.debug('handle_pl() < "%s"', s)
    res = re.match(r'^v?(\d+)\.(\d+)\.(\d+)-pl(\d+)', s)
    if not res:
        logger.debug('handle_pl() -> ""')
        return ''
    major, minor, mm, pl = res.groups()
    ret = f'{major}.{minor}.{mm}.{pl}'
    logger.debug('handle_pl() -> "%s"', ret)
    return ret



"""String utility functions."""
from __future__ import annotations

from functools import cache
from typing import Literal
from urllib.parse import urlparse
import logging
import re

log = logging.getLogger(__name__)
# From parse-package-name
# https://github.com/egoist/parse-package-name/blob/main/src/index.ts
RE_SCOPED = r'^(@[^\/]+\/[^@\/]+)(?:@([^\/]+))?(\/.*)?$'
"""Regular expression for a Node.js scoped package name."""
RE_NON_SCOPED = r'^([^@\/]+)(?:@([^\/]+))?(\/.*)?$'
"""Regular expression for a Node.js non-scoped package name."""


@cache
def dotize(s: str) -> str:
    """Convert dashes and underscores in a string to full stops."""
    ret = s.replace('-', '.').replace('_', '.')
    log.debug('dotize(): %s -> %s', s, ret)
    return ret


@cache
def is_sha(url: str) -> Literal[40, 7, 0]:
    """
    Extract the last part of a URL and checks if it is a valid SHA-1 hash.

    Parameters
    ----------
    url : str
        The input URL string.

    Returns
    -------
    int
        Returns ``7`` if it's a short SHA, ``40`` if it's a full SHA, ``0`` otherwise.
    """
    last_part = urlparse(url).path.rsplit('/', 1)[-1] if '/' in url else url

    if re.match(r'^[0-9a-f]{40}', last_part):
        return 40
    if re.match(r'^[0-9a-f]{7}', last_part):
        return 7
    return 0


def extract_sha(text: str) -> str:
    """
    Extract the first valid SHA-1 hash (7 or 40 characters) found in the given string.

    Parameters
    ----------
    text : str
        The input string to search for a SHA-1 hash.

    Returns
    -------
    str | None
        The SHA-1 hash if found, otherwise empty string.
    """
    match = re.search(r'\b(?:[0-9a-f]{7}|[0-9a-f]{40})\b', text)

    return match.group(0) if match else ''


class InvalidPackageName(ValueError):
    """Raised when a package name is invalid."""
    def __init__(self, pkg: str) -> None:
        super().__init__(f'Invalid package name: {pkg}.')


def parse_npm_package_name(s: str) -> tuple[str, str | None, str | None]:
    """
    Parse a Node.js package name into its components.

    Parameters.
    ----------
    s : str
        The package name to parse.

    Returns
    -------
    tuple[str, str | None, str | None]
        A tuple containing the scope, name, and version of the package.

    Raises
    ------
    InvalidPackageName
        If the package name does not match the expected format.
    """
    if not (m := re.match(RE_SCOPED, s) or re.match(RE_NON_SCOPED, s)):
        raise InvalidPackageName(s)
    return m[1], m[2], m[3]


@cache
def prefix_v(s: str) -> str:
    """Prefix a version string with 'v'."""
    return f'v{s}'


@cache
def dash_to_underscore(s: str) -> str:
    """Convert dashes in a string to underscores."""
    return s.replace('-', '_')



from __future__ import annotations

from typing import TYPE_CHECKING

from livecheck.special.pecl import (
    get_latest_pecl_metadata,
    get_latest_pecl_package,
    get_latest_pecl_package2,
)
import pytest

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_get_latest_pecl_package_removes_prefix_and_calls_helper(mocker: MockerFixture) -> None:
    ebuild = 'dev-php/pecl-foo-1.0.0'
    settings = mocker.Mock()
    mocker.patch('livecheck.special.pecl.catpkg_catpkgsplit',
                 return_value=('dev-php', None, 'pecl-foo', None))
    helper = mocker.patch('livecheck.special.pecl.get_latest_pecl_package2', return_value='1.2.3')
    result = get_latest_pecl_package(ebuild, settings)
    helper.assert_called_once_with('foo', ebuild, settings)
    assert result == '1.2.3'


def test_get_latest_pecl_package_no_prefix(mocker: MockerFixture) -> None:
    ebuild = 'dev-php/foo-1.0.0'
    settings = mocker.Mock()
    mocker.patch('livecheck.special.pecl.catpkg_catpkgsplit',
                 return_value=('dev-php', None, 'foo', None))
    helper = mocker.patch('livecheck.special.pecl.get_latest_pecl_package2', return_value='2.0.0')

    result = get_latest_pecl_package(ebuild, settings)
    helper.assert_called_once_with('foo', ebuild, settings)
    assert result == '2.0.0'


def test_get_latest_pecl_package2_returns_latest_stable(mocker: MockerFixture) -> None:
    ebuild = 'dev-php/foo-1.0.0'
    settings = mocker.Mock()
    settings.is_devel.return_value = False
    mocker.patch('livecheck.special.pecl.catpkg_catpkgsplit',
                 return_value=('dev-php', None, 'foo', None))
    xml = """
    <a xmlns="http://pear.php.net/dtd/rest.allreleases">
      <r><v>1.0.0</v><s>stable</s></r>
      <r><v>1.1.0</v><s>beta</s></r>
      <r><v>2.0.0</v><s>stable</s></r>
    </a>
    """
    mock_get_content = mocker.patch('livecheck.special.pecl.get_content',
                                    return_value=mocker.Mock(text=xml))
    mock_get_last_version = mocker.patch('livecheck.special.pecl.get_last_version',
                                         return_value={'version': '2.0.0'})

    result = get_latest_pecl_package2('foo', ebuild, settings)

    assert result == '2.0.0'
    mock_get_content.assert_called_once()
    mock_get_last_version.assert_called_once()


def test_get_latest_pecl_package2_returns_empty_on_no_content(mocker: MockerFixture) -> None:
    ebuild = 'dev-php/foo-1.0.0'
    settings = mocker.Mock()
    mocker.patch('livecheck.special.pecl.catpkg_catpkgsplit',
                 return_value=('dev-php', None, 'foo', None))
    mocker.patch('livecheck.special.pecl.get_content', return_value=None)

    result = get_latest_pecl_package2('foo', ebuild, settings)
    assert not result


def test_get_latest_pecl_package2_returns_empty_on_no_last_version(mocker: MockerFixture) -> None:
    ebuild = 'dev-php/foo-1.0.0'
    settings = mocker.Mock()
    mocker.patch('livecheck.special.pecl.catpkg_catpkgsplit',
                 return_value=('dev-php', None, 'foo', None))
    xml = """
    <a xmlns="http://pear.php.net/dtd/rest.allreleases">
      <r><v>1.0.0</v><s>stable</s></r>
    </a>
    """
    mocker.patch('livecheck.special.pecl.get_content', return_value=mocker.Mock(text=xml))
    mocker.patch('livecheck.special.pecl.get_last_version', return_value=None)

    get_latest_pecl_package2('foo', ebuild, settings)


def test_get_latest_pecl_package2_filters_only_stable_when_not_devel(mocker: MockerFixture) -> None:
    ebuild = 'dev-php/foo-1.0.0'
    settings = mocker.Mock()
    settings.is_devel.return_value = False
    mocker.patch('livecheck.special.pecl.catpkg_catpkgsplit',
                 return_value=('dev-php', None, 'foo', None))
    xml = """
    <a xmlns="http://pear.php.net/dtd/rest.allreleases">
        <r><v>1.0.0</v><s>stable</s></r>
        <r><v>1.1.0</v><s>beta</s></r>
        <r><v>2.0.0</v><s>stable</s></r>
        <r><v>3.0.0</v><s>alpha</s></r>
    </a>
    """
    mocker.patch('livecheck.special.pecl.get_content', return_value=mocker.Mock(text=xml))
    mock_get_last_version = mocker.patch('livecheck.special.pecl.get_last_version',
                                         return_value={'version': '2.0.0'})

    result = get_latest_pecl_package2('foo', ebuild, settings)

    assert result == '2.0.0'
    mock_get_last_version.assert_called_once()
    # Only stable versions should be considered
    args, _kwargs = mock_get_last_version.call_args
    tags = args[0]
    assert all(tag['tag'] in {'1.0.0', '2.0.0'} for tag in tags)


def test_get_latest_pecl_package2_includes_all_when_devel(mocker: MockerFixture) -> None:
    ebuild = 'dev-php/foo-1.0.0'
    settings = mocker.Mock()
    settings.is_devel.return_value = True
    mocker.patch('livecheck.special.pecl.catpkg_catpkgsplit',
                 return_value=('dev-php', None, 'foo', None))
    xml = """
    <a xmlns="http://pear.php.net/dtd/rest.allreleases">
        <r><v>1.0.0</v><s>stable</s></r>
        <r><v>1.1.0</v><s>beta</s></r>
        <r><v>2.0.0</v><s>stable</s></r>
        <r><v>3.0.0</v><s>alpha</s></r>
    </a>
    """
    mocker.patch('livecheck.special.pecl.get_content', return_value=mocker.Mock(text=xml))
    mock_get_last_version = mocker.patch('livecheck.special.pecl.get_last_version',
                                         return_value={'version': '3.0.0'})

    result = get_latest_pecl_package2('foo', ebuild, settings)

    assert result == '3.0.0'
    mock_get_last_version.assert_called_once()
    args, _kwargs = mock_get_last_version.call_args
    tags = args[0]
    # All versions should be included
    assert {tag['tag'] for tag in tags} == {'1.0.0', '1.1.0', '2.0.0', '3.0.0'}


def test_get_latest_pecl_package2_returns_empty_if_no_releases(mocker: MockerFixture) -> None:
    ebuild = 'dev-php/foo-1.0.0'
    settings = mocker.Mock()
    settings.is_devel.return_value = False
    mocker.patch('livecheck.special.pecl.catpkg_catpkgsplit',
                 return_value=('dev-php', None, 'foo', None))
    xml = """
    <a xmlns="http://pear.php.net/dtd/rest.allreleases">
    </a>
    """
    mocker.patch('livecheck.special.pecl.get_content', return_value=mocker.Mock(text=xml))
    mock_get_last_version = mocker.patch('livecheck.special.pecl.get_last_version',
                                         return_value=None)

    result = get_latest_pecl_package2('foo', ebuild, settings)
    assert not result
    mock_get_last_version.assert_called_once()


def test_get_latest_pecl_metadata_calls_helper_with_correct_args(mocker: MockerFixture) -> None:
    remote = 'foo'
    ebuild = 'dev-php/foo-1.0.0'
    settings = mocker.Mock()
    mock_helper = mocker.patch('livecheck.special.pecl.get_latest_pecl_package2',
                               return_value='9.9.9')
    result = get_latest_pecl_metadata(remote, ebuild, settings)
    mock_helper.assert_called_once_with(remote, ebuild, settings)
    assert result == '9.9.9'


def test_get_latest_pecl_metadata_returns_empty_when_helper_returns_empty(
        mocker: MockerFixture) -> None:
    remote = 'foo'
    ebuild = 'dev-php/foo-1.0.0'
    settings = mocker.Mock()
    mocker.patch('livecheck.special.pecl.get_latest_pecl_package2', return_value='')
    result = get_latest_pecl_metadata(remote, ebuild, settings)
    assert not result


def test_get_latest_pecl_metadata_passes_through_exceptions(mocker: MockerFixture) -> None:
    remote = 'foo'
    ebuild = 'dev-php/foo-1.0.0'
    settings = mocker.Mock()
    mocker.patch('livecheck.special.pecl.get_latest_pecl_package2', side_effect=ValueError('fail'))
    with pytest.raises(ValueError, match='fail'):
        get_latest_pecl_metadata(remote, ebuild, settings)



"""Utilities for requests module."""
from __future__ import annotations

from dataclasses import dataclass
from email.utils import parsedate_to_datetime
from functools import cache
from http import HTTPStatus
from urllib.parse import urlparse
import hashlib
import logging

import requests

from .credentials import get_api_credentials

log = logging.getLogger(__name__)


@dataclass
class TextDataResponse:
    """Used for data URI responses."""
    text: str
    """Text content."""
    status_code: int = HTTPStatus.OK
    """Defaults to 200 OK."""
    def raise_for_status(self) -> None:
        """Do nothing."""


@cache
def session_init(module: str) -> requests.Session:
    """Create a session."""
    session = requests.Session()
    if module == 'github':
        token = get_api_credentials('github.com')
        if token:
            session.headers['Authorization'] = f'Bearer {token}'
        session.headers['Accept'] = 'application/vnd.github.v3+json'
    elif module == 'xml':
        session.headers['Accept'] = 'application/xml'
    elif module == 'json':
        session.headers['Accept'] = 'application/json'
    elif module == 'gitlab':
        token = get_api_credentials('gitlab.com')
        if token:
            session.headers['Authorization'] = f'Bearer {token}'
        session.headers['Accept'] = 'application/json'
    elif module == 'bitbucket':
        token = get_api_credentials('bitbucket.org')
        if token:
            session.headers['Authorization'] = f'Bearer {token}'
        session.headers['Accept'] = 'application/json'
    session.headers['timeout'] = '30'
    return session


def get_content(url: str) -> requests.Response:
    """"Fetch content from a URL."""
    parsed_uri = urlparse(url)
    log.debug('Fetching %s', url)

    if parsed_uri.scheme == 'mirror':
        # If the URL is a mirror, we need to handle it differently
        # This is a placeholder for the actual implementation
        log.debug('Handling mirror:// protocol for `%s`.', url)
        response = requests.Response()
        response.status_code = HTTPStatus.NOT_IMPLEMENTED
        return response

    if parsed_uri.hostname == 'api.github.com':
        session = session_init('github')
    elif parsed_uri.hostname == 'api.gitlab.com':
        session = session_init('gitlab')
    elif parsed_uri.hostname == 'api.bitbucket.org':
        session = session_init('bitbucket')
    elif parsed_uri.hostname == 'repology.org':
        session = session_init('json')
        session.headers['User-Agent'] = 'DistroWatch'
    elif url.endswith(('.atom', '.xml')):
        session = session_init('xml')
    elif url.endswith('json'):
        session = session_init('json')
    else:
        session = session_init('')

    r: TextDataResponse | requests.Response
    try:
        r = session.get(url)
    except requests.RequestException:
        log.exception('Caught error {e} attempting to fetch {url}')
        r = requests.Response()
        r.status_code = HTTPStatus.SERVICE_UNAVAILABLE
        return r
    if r.status_code not in {
            HTTPStatus.OK, HTTPStatus.CREATED, HTTPStatus.ACCEPTED, HTTPStatus.PARTIAL_CONTENT,
            HTTPStatus.MOVED_PERMANENTLY, HTTPStatus.FOUND, HTTPStatus.TEMPORARY_REDIRECT,
            HTTPStatus.PERMANENT_REDIRECT
    }:
        log.error('Error fetching %s. Status code: %d', url, r.status_code)
    elif not r.text:
        log.warning('Empty response for %s.', url)

    return r


@cache
def hash_url(url: str) -> tuple[str, str, int]:
    """Hash the content of a URL using BLAKE2b and SHA-512."""
    h_blake2b = hashlib.blake2b()
    h_sha512 = hashlib.sha512()
    size = 0
    try:
        with requests.get(url, stream=True, timeout=30) as r:
            r.raise_for_status()
            for chunk in r.iter_content(chunk_size=8192):
                if chunk:
                    h_blake2b.update(chunk)
                    h_sha512.update(chunk)
                    size += len(chunk)
        return h_blake2b.hexdigest(), h_sha512.hexdigest(), size
    except requests.RequestException:
        log.exception('Error hashing URL %s.', url)

    return '', '', 0


@cache
def get_last_modified(url: str) -> str:
    """Get the last modified date of a URL."""
    try:
        with requests.head(url, timeout=30) as r:
            r.raise_for_status()
            if last_modified := r.headers.get('last-modified'):
                return parsedate_to_datetime(last_modified).strftime('%Y%m%d')

    except requests.RequestException:
        log.exception('Error fetching last modified header for %s.', url)

    return ''



"""Constants."""
from __future__ import annotations

from typing import TYPE_CHECKING

from .utils import prefix_v

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping

__all__ = ('RSS_NS', 'SUBMODULES', 'TAG_NAME_FUNCTIONS')

RSS_NS = {'': 'http://www.w3.org/2005/Atom'}
"""
Namespace for RSS feeds.

:meta hide-value:
"""
SUBMODULES: Mapping[str, set[str | tuple[str, str]]] = {
    'app-misc/tasksh': {'src/libshared'},
    'app-pda/tsschecker': {'external/jssy'},
    'games-emulation/citra': {
        'externals/dynarmic',
        'externals/fmt',
        'externals/lodepng/lodepng',
        'externals/sirit',
        'externals/soundtouch',
        'externals/xbyak',
    },
    'games-emulation/play': {
        'deps/CodeGen',
        'deps/Dependencies',
        'deps/Framework',
        'deps/Nuanceur',
        'deps/libchdr',
    },
    'games-emulation/rpcs3': {
        '3rdparty/asmjit/asmjit',
        '3rdparty/hidapi/hidapi',
        '3rdparty/yaml-cpp/yaml-cpp',
        '3rdparty/SoundTouch/soundtouch',
        '3rdparty/llvm/llvm',
    },
    'games-emulation/sudachi': {
        'externals/SDL',
        'externals/cpp-httplib',
        'externals/cpp-jwt',
        'externals/dynarmic',
        'externals/ffmpeg/ffmpeg',
        'externals/mbedtls',
        'externals/simpleini',
        'externals/sirit',
        'externals/xbyak',
    },
    'games-emulation/vita3k': {
        'external/LibAtrac9',
        'external/SPIRV-Cross',
        'external/VulkanMemoryAllocator-Hpp',
        'external/better-enums',
        'external/crypto-algorithms',
        'external/dlmalloc',
        'external/dynarmic',
        'external/fmt',
        'external/imgui',
        'external/imgui_club',
        'external/libfat16',
        'external/printf',
        'external/spdlog',
        'external/stb',
        'external/unicorn',
        'external/vita-toolchain',
    },
    'games-emulation/yuzu': {
        'externals/SDL',
        'externals/cpp-jwt',
        'externals/dynarmic',
        'externals/mbedtls',
        'externals/sirit',
        ('externals/cpp-httplib', 'HTTPLIB_SHA'),
    },
    'media-sound/sony-headphones-client': {'Client/imgui'}
}
"""
Currently, submodule handling is hard-coded for specific packages.

:meta hide-value:
"""
TAG_NAME_FUNCTIONS: Mapping[str, Callable[[str], str]] = {
    'app-misc/tasksh': prefix_v,
    'games-emulation/rpcs3': prefix_v,
    'games-emulation/xemu': prefix_v,
    'games-emulation/yuzu': lambda x: f'mainline-{x.replace(".", "-")}',
    'media-sound/sony-headphones-client': prefix_v,
}
"""
Functions to prefix tag names for specific packages.

:meta hide-value:
"""
